# -*- coding: utf-8 -*-
"""The pinfo CLI tool."""

from __future__ import unicode_literals

import argparse
import collections
import json
import os
import uuid

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.cli import logger
from plaso.cli import tool_options
from plaso.cli import tools
from plaso.cli import views
from plaso.cli.helpers import manager as helpers_manager
from plaso.engine import knowledge_base
from plaso.lib import definitions
from plaso.lib import errors
from plaso.lib import loggers
from plaso.serializer import json_serializer
from plaso.storage import factory as storage_factory


class PinfoTool(
    tools.CLITool,
    tool_options.StorageFileOptions):
  """Pinfo CLI tool."""

  NAME = 'pinfo'
  DESCRIPTION = (
      'Shows information about a Plaso storage file, for example how it was '
      'collected, what information was extracted from a source, etc.')

  _INDENTATION_LEVEL = 8

  def __init__(self, input_reader=None, output_writer=None):
    """Initializes the CLI tool object.

    Args:
      input_reader (Optional[InputReader]): input reader, where None indicates
          that the stdin input reader should be used.
      output_writer (Optional[OutputWriter]): output writer, where None
          indicates that the stdout output writer should be used.
    """
    super(PinfoTool, self).__init__(
        input_reader=input_reader, output_writer=output_writer)
    self._compare_storage_file_path = None
    self._output_filename = None
    self._output_format = None
    self._process_memory_limit = None
    self._storage_file_path = None
    self._verbose = False

    self.compare_storage_information = False

  def _CalculateStorageCounters(self, storage_reader):
    """Calculates the counters of the entire storage.

    Args:
      storage_reader (StorageReader): storage reader.

    Returns:
      dict[str, collections.Counter]: storage counters.
    """
    analysis_reports_counter = collections.Counter()
    analysis_reports_counter_error = False
    event_labels_counter = collections.Counter()
    event_labels_counter_error = False
    parsers_counter = collections.Counter()
    parsers_counter_error = False

    for session in storage_reader.GetSessions():
      if isinstance(session.analysis_reports_counter, collections.Counter):
        analysis_reports_counter += session.analysis_reports_counter
      else:
        analysis_reports_counter_error = True

      if isinstance(session.event_labels_counter, collections.Counter):
        event_labels_counter += session.event_labels_counter
      else:
        event_labels_counter_error = True

      if isinstance(session.parsers_counter, collections.Counter):
        parsers_counter += session.parsers_counter
      else:
        parsers_counter_error = True

    storage_counters = {}

    warnings_by_path_spec = collections.Counter()
    warnings_by_parser_chain = collections.Counter()

    for warning in list(storage_reader.GetWarnings()):
      warnings_by_path_spec[warning.path_spec.comparable] += 1
      warnings_by_parser_chain[warning.parser_chain] += 1

    storage_counters['warnings_by_path_spec'] = warnings_by_path_spec
    storage_counters['warnings_by_parser_chain'] = warnings_by_parser_chain

    if not analysis_reports_counter_error:
      storage_counters['analysis_reports'] = analysis_reports_counter

    if not event_labels_counter_error:
      storage_counters['event_labels'] = event_labels_counter

    if not parsers_counter_error:
      storage_counters['parsers'] = parsers_counter

    return storage_counters

  def _CompareCounter(self, counter, compare_counter):
    """Compares two counters.

    Args:
      counter (collections.Counter): counter.
      compare_counter (collections.Counter): counter to compare against.

    Returns:
      dict[str, tuple[int, int]]: mismatching results per key.
    """
    keys = set(counter.keys())
    keys.union(compare_counter.keys())

    differences = {}
    for key in keys:
      value = counter.get(key, 0)
      compare_value = compare_counter.get(key, 0)
      if value != compare_value:
        differences[key] = (value, compare_value)

    return differences

  def _PrintCounterDifferences(
      self, differences, column_names=None, reverse=False, title=None):
    """Prints the counter differences.

    Args:
      differences (dict[str, tuple[int, int]]): mismatching results per key.
      column_names (Optional[list[str]]): column names.
      reverse (Optional[bool]): True if the key and values of differences
          should be printed in reverse order.
      title (Optional[str]): title.
    """
    # TODO: add support for 3 column table?
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=column_names, title=title)

    for key, value in sorted(differences.items()):
      value_string = '{0:d} ({1:d})'.format(value[0], value[1])
      if reverse:
        table_view.AddRow([value_string, key])
      else:
        table_view.AddRow([key, value_string])

    table_view.Write(self._output_writer)
    self._output_writer.Write('\n')

  def _CompareStores(self, storage_reader, compare_storage_reader):
    """Compares the contents of two stores.

    Args:
      storage_reader (StorageReader): storage reader.
      compare_storage_reader (StorageReader): storage to compare against.

    Returns:
      bool: True if the content of the stores is identical.
    """
    stores_are_identical = True

    storage_counters = self._CalculateStorageCounters(storage_reader)
    compare_storage_counters = self._CalculateStorageCounters(
        compare_storage_reader)

    # Compare number of events.
    parsers_counter = storage_counters.get('parsers', collections.Counter())
    compare_parsers_counter = compare_storage_counters.get(
        'parsers', collections.Counter())
    differences = self._CompareCounter(parsers_counter, compare_parsers_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences,
          column_names=['Parser (plugin) name', 'Number of events'],
          title='Events generated per parser')

    # Compare warnings by parser chain.
    warnings_counter = storage_counters.get(
        'warnings_by_parser_chain', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'warnings_by_parser_chain', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences,
          column_names=['Parser (plugin) name', 'Number of warnings'],
          title='Warnings generated per parser')

    # Compare warnings by path specification
    warnings_counter = storage_counters.get(
        'warnings_by_path_spec', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'warnings_by_path_spec', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Number of warnings', 'Pathspec'],
          reverse=True, title='Pathspecs with most warnings')

    # Compare event labels.
    labels_counter = storage_counters.get('event_labels', collections.Counter())
    compare_labels_counter = compare_storage_counters.get(
        'event_labels', collections.Counter())
    differences = self._CompareCounter(labels_counter, compare_labels_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Label', 'Number of event tags'],
          title='Event tags generated per label')

    # Compare analysis reports.
    reports_counter = storage_counters.get(
        'analysis_reports', collections.Counter())
    compare_reports_counter = compare_storage_counters.get(
        'analysis_reports', collections.Counter())
    differences = self._CompareCounter(reports_counter, compare_reports_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Plugin name', 'Number of reports'],
          title='Reports generated per plugin')

    return stores_are_identical

  def _PrintAnalysisReportCounter(
      self, analysis_reports_counter, session_identifier=None):
    """Prints the analysis reports counter.

    Args:
      analysis_reports_counter (collections.Counter): number of analysis
          reports per analysis plugin.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if not analysis_reports_counter:
      return

    title = 'Reports generated per plugin'
    if session_identifier:
      title = '{0:s}: {1:s}'.format(title, session_identifier)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Plugin name', 'Number of reports'], title=title)

    for key, value in sorted(analysis_reports_counter.items()):
      if key != 'total':
        table_view.AddRow([key, value])

    try:
      total = analysis_reports_counter['total']
    except KeyError:
      total = 'N/A'

    table_view.AddRow(['Total', total])

    table_view.Write(self._output_writer)

  def _PrintAnalysisReportsDetails(self, storage_reader):
    """Prints the details of the analysis reports.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    if not storage_reader.HasAnalysisReports():
      self._output_writer.Write('No analysis reports stored.\n\n')
      return

    for index, analysis_report in enumerate(
        storage_reader.GetAnalysisReports()):
      title = 'Analysis report: {0:d}'.format(index)
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, title=title)

      table_view.AddRow(['String', analysis_report.GetString()])

      table_view.Write(self._output_writer)

  def _PrintWarningCounters(self, storage_counters):
    """Prints a summary of the warnings.

    Args:
      storage_counters (dict): storage counters.
    """
    warnings_by_pathspec = storage_counters.get('warnings_by_path_spec', {})
    warnings_by_parser_chain = storage_counters.get(
        'warnings_by_parser_chain', {})
    if not warnings_by_parser_chain:
      self._output_writer.Write('No warnings stored.\n\n')
      return

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Parser (plugin) name', 'Number of warnings'],
        title='Warnings generated per parser')
    for parser_chain, count in warnings_by_parser_chain.items():
      parser_chain = parser_chain or '<No parser>'
      table_view.AddRow([parser_chain, '{0:d}'.format(count)])
    table_view.Write(self._output_writer)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Number of warnings', 'Pathspec'],
        title='Pathspecs with most warnings')

    top_pathspecs = warnings_by_pathspec.most_common(10)
    for pathspec, count in top_pathspecs:
      for path_index, line in enumerate(pathspec.split('\n')):
        if not line:
          continue

        if path_index == 0:
          table_view.AddRow(['{0:d}'.format(count), line])
        else:
          table_view.AddRow(['', line])

    table_view.Write(self._output_writer)

  def _PrintWarningsDetails(self, storage):
    """Prints the details of the warnings.

    Args:
      storage (BaseStore): storage.
    """
    if not storage.HasWarnings():
      self._output_writer.Write('No warnings stored.\n\n')
      return

    for index, warning in enumerate(storage.GetWarnings()):
      title = 'Warning: {0:d}'.format(index)
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, title=title)

      table_view.AddRow(['Message', warning.message])
      table_view.AddRow(['Parser chain', warning.parser_chain])

      path_specification = warning.path_spec.comparable
      for path_index, line in enumerate(path_specification.split('\n')):
        if not line:
          continue

        if path_index == 0:
          table_view.AddRow(['Path specification', line])
        else:
          table_view.AddRow(['', line])

      table_view.Write(self._output_writer)

  def _PrintEventLabelsCounter(
      self, event_labels_counter, session_identifier=None):
    """Prints the event labels counter.

    Args:
      event_labels_counter (collections.Counter): number of event tags per
          label.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if not event_labels_counter:
      return

    title = 'Event tags generated per label'
    if session_identifier:
      title = '{0:s}: {1:s}'.format(title, session_identifier)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Label', 'Number of event tags'], title=title)

    for key, value in sorted(event_labels_counter.items()):
      if key != 'total':
        table_view.AddRow([key, value])

    try:
      total = event_labels_counter['total']
    except KeyError:
      total = 'N/A'

    table_view.AddRow(['Total', total])

    table_view.Write(self._output_writer)

  def _PrintParsersCounter(self, parsers_counter, session_identifier=None):
    """Prints the parsers counter

    Args:
      parsers_counter (collections.Counter): number of events per parser or
          parser plugin.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if not parsers_counter:
      self._output_writer.Write('No events stored.\n\n')
      return

    title = 'Events generated per parser'
    if session_identifier:
      title = '{0:s}: {1:s}'.format(title, session_identifier)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Parser (plugin) name', 'Number of events'],
        title=title)

    for key, value in sorted(parsers_counter.items()):
      if key != 'total':
        table_view.AddRow([key, value])

    table_view.AddRow(['Total', parsers_counter['total']])

    table_view.Write(self._output_writer)

  def _PrintPreprocessingInformation(
      self, storage_reader, session_identifier=None):
    """Prints the details of the preprocessing information.

    Args:
      storage_reader (StorageReader): storage reader.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    knowledge_base_object = knowledge_base.KnowledgeBase()

    storage_reader.ReadPreprocessingInformation(knowledge_base_object)

    lookup_identifier = session_identifier
    if lookup_identifier:
      # The knowledge base requires the session identifier to be formatted in
      # hexadecimal representation.
      lookup_identifier = lookup_identifier.replace('-', '')

    system_configuration = knowledge_base_object.GetSystemConfigurationArtifact(
        session_identifier=lookup_identifier)
    if not system_configuration:
      return

    title = 'System configuration'
    if session_identifier:
      title = '{0:s}: {1:s}'.format(title, session_identifier)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title=title)

    hostname = 'N/A'
    if system_configuration.hostname:
      hostname = system_configuration.hostname.name

    operating_system = system_configuration.operating_system or 'N/A'
    operating_system_product = (
        system_configuration.operating_system_product or 'N/A')
    operating_system_version = (
        system_configuration.operating_system_version or 'N/A')
    code_page = system_configuration.code_page or 'N/A'
    keyboard_layout = system_configuration.keyboard_layout or 'N/A'
    time_zone = system_configuration.time_zone or 'N/A'

    table_view.AddRow(['Hostname', hostname])
    table_view.AddRow(['Operating system', operating_system])
    table_view.AddRow(['Operating system product', operating_system_product])
    table_view.AddRow(['Operating system version', operating_system_version])
    table_view.AddRow(['Code page', code_page])
    table_view.AddRow(['Keyboard layout', keyboard_layout])
    table_view.AddRow(['Time zone', time_zone])

    table_view.Write(self._output_writer)

    title = 'Available time zones'
    if session_identifier:
      title = '{0:s}: {1:s}'.format(title, session_identifier)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Name', ''], title=title)

    for time_zone in system_configuration.available_time_zones:
      table_view.AddRow([time_zone.name, ''])

    table_view.Write(self._output_writer)

    title = 'User accounts'
    if session_identifier:
      title = '{0:s}: {1:s}'.format(title, session_identifier)

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Username', 'User directory'], title=title)

    for user_account in system_configuration.user_accounts:
      table_view.AddRow([
          user_account.username, user_account.user_directory])

    table_view.Write(self._output_writer)

  def _PrintSessionsDetails(self, storage_reader):
    """Prints the details of the sessions.

    Args:
      storage_reader (BaseStore): storage.
    """
    for session in storage_reader.GetSessions():
      session_identifier = uuid.UUID(hex=session.identifier)
      session_identifier = '{0!s}'.format(session_identifier)

      start_time = 'N/A'
      if session.start_time is not None:
        date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
            timestamp=session.start_time)
        start_time = date_time.CopyToDateTimeStringISO8601()

      completion_time = 'N/A'
      if session.completion_time is not None:
        date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
            timestamp=session.completion_time)
        completion_time = date_time.CopyToDateTimeStringISO8601()

      enabled_parser_names = 'N/A'
      if session.enabled_parser_names:
        enabled_parser_names = ', '.join(sorted(session.enabled_parser_names))

      command_line_arguments = session.command_line_arguments or 'N/A'
      parser_filter_expression = session.parser_filter_expression or 'N/A'
      preferred_encoding = session.preferred_encoding or 'N/A'

      if session.artifact_filters:
        artifact_filters_string = ', '.join(session.artifact_filters)
      else:
        artifact_filters_string = 'N/A'
      filter_file = session.filter_file or 'N/A'
      number_of_event_sources = storage_reader.GetNumberOfEventSources()

      title = 'Session: {0:s}'.format(session_identifier)
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, title=title)

      table_view.AddRow(['Start time', start_time])
      table_view.AddRow(['Completion time', completion_time])
      table_view.AddRow(['Product name', session.product_name])
      table_view.AddRow(['Product version', session.product_version])
      table_view.AddRow(['Command line arguments', command_line_arguments])
      table_view.AddRow(['Parser filter expression', parser_filter_expression])
      table_view.AddRow(['Enabled parser and plugins', enabled_parser_names])
      table_view.AddRow(['Preferred encoding', preferred_encoding])
      table_view.AddRow(['Debug mode', session.debug_mode])
      table_view.AddRow(['Artifact filters', artifact_filters_string])
      table_view.AddRow(['Filter file', filter_file])
      table_view.AddRow(['Number of event sources', number_of_event_sources])

      table_view.Write(self._output_writer)

      if self._verbose:
        self._PrintPreprocessingInformation(
            storage_reader, session_identifier=session_identifier)

        self._PrintParsersCounter(
            session.parsers_counter, session_identifier=session_identifier)

        self._PrintAnalysisReportCounter(
            session.analysis_reports_counter,
            session_identifier=session_identifier)

        self._PrintEventLabelsCounter(
            session.event_labels_counter,
            session_identifier=session_identifier)

  def _PrintSessionsOverview(self, storage_reader):
    """Prints a sessions overview.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title='Sessions')

    for session in storage_reader.GetSessions():
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=session.start_time)
      start_time = date_time.CopyToDateTimeStringISO8601()

      session_identifier = uuid.UUID(hex=session.identifier)
      session_identifier = '{0!s}'.format(session_identifier)
      table_view.AddRow([session_identifier, start_time])

    table_view.Write(self._output_writer)

  def _PrintStorageInformationAsText(self, storage_reader):
    """Prints information about the store as human-readable text.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    storage_type = storage_reader.GetStorageType()
    serialization_format = storage_reader.GetSerializationFormat()
    format_version = storage_reader.GetFormatVersion()

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title='Plaso Storage Information')
    table_view.AddRow(['Filename', os.path.basename(self._storage_file_path)])
    table_view.AddRow(['Format version', format_version])
    table_view.AddRow(['Serialization format', serialization_format])
    table_view.Write(self._output_writer)

    if storage_type == definitions.STORAGE_TYPE_SESSION:
      self._PrintSessionsOverview(storage_reader)
      self._PrintSessionsDetails(storage_reader)

      storage_counters = self._CalculateStorageCounters(storage_reader)

      if 'parsers' not in storage_counters:
        self._output_writer.Write(
            'Unable to determine number of events generated per parser.\n')
      else:
        self._PrintParsersCounter(storage_counters['parsers'])

      if 'analysis_reports' not in storage_counters:
        self._output_writer.Write(
            'Unable to determine number of reports generated per plugin.\n')
      else:
        self._PrintAnalysisReportCounter(storage_counters['analysis_reports'])

      if 'event_labels' not in storage_counters:
        self._output_writer.Write(
            'Unable to determine number of event tags generated per label.\n')
      else:
        self._PrintEventLabelsCounter(storage_counters['event_labels'])

      self._PrintWarningCounters(storage_counters)

      if self._verbose:
        self._PrintWarningsDetails(storage_reader)

      self._PrintAnalysisReportsDetails(storage_reader)

    elif storage_type == definitions.STORAGE_TYPE_TASK:
      self._PrintTasksInformation(storage_reader)

  def _PrintStorageInformationAsJSON(self, storage_reader):
    """Writes a summary of sessions as machine-readable JSON.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    serializer = json_serializer.JSONAttributeContainerSerializer
    storage_counters = self._CalculateStorageCounters(storage_reader)
    storage_counters_json = json.dumps(storage_counters)
    self._output_writer.Write('{')
    self._output_writer.Write('"storage_counters": {0:s}'.format(
        storage_counters_json))
    self._output_writer.Write(',\n')
    self._output_writer.Write(' "sessions": {')
    for index, session in enumerate(storage_reader.GetSessions()):
      json_string = serializer.WriteSerialized(session)
      if index != 0:
        self._output_writer.Write(',\n')
      self._output_writer.Write('"session_{0:s}": {1:s} '.format(
          session.identifier, json_string))
    self._output_writer.Write('}}')

  def _PrintTasksInformation(self, storage_reader):
    """Prints information about the tasks.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title='Tasks')

    for task_start, _ in storage_reader.GetSessions():
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=task_start.timestamp)
      start_time = date_time.CopyToDateTimeStringISO8601()

      task_identifier = uuid.UUID(hex=task_start.identifier)
      task_identifier = '{0!s}'.format(task_identifier)
      table_view.AddRow([task_identifier, start_time])

    table_view.Write(self._output_writer)

  def CompareStores(self):
    """Compares the contents of two stores.

    Returns:
      bool: True if the content of the stores is identical.
    """
    storage_reader = storage_factory.StorageFactory.CreateStorageReaderForFile(
        self._storage_file_path)
    if not storage_reader:
      logger.error(
          'Format of storage file: {0:s} not supported'.format(
              self._storage_file_path))
      return False

    compare_storage_reader = (
        storage_factory.StorageFactory.CreateStorageReaderForFile(
            self._compare_storage_file_path))
    if not compare_storage_reader:
      logger.error(
          'Format of storage file: {0:s} not supported'.format(
              self._compare_storage_file_path))
      return False

    try:
      result = self._CompareStores(storage_reader, compare_storage_reader)

    finally:
      compare_storage_reader.Close()
      storage_reader.Close()

    if result:
      self._output_writer.Write('Storage files are identical.\n')
    else:
      self._output_writer.Write('Storage files are different.\n')

    return result

  def ParseArguments(self, arguments):
    """Parses the command line arguments.

    Args:
      arguments (list[str]): command line arguments.

    Returns:
      bool: True if the arguments were successfully parsed.
    """
    loggers.ConfigureLogging()

    argument_parser = argparse.ArgumentParser(
        description=self.DESCRIPTION, add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    self.AddBasicOptions(argument_parser)

    argument_helper_names = ['storage_file']
    if self._CanEnforceProcessMemoryLimit():
      argument_helper_names.append('process_resources')
    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        argument_parser, names=argument_helper_names)

    argument_parser.add_argument(
        '--compare', dest='compare_storage_file', type=str,
        action='store', default='', metavar='STORAGE_FILE', help=(
            'The path of the storage file to compare against.'))

    argument_parser.add_argument(
        '--output_format', '--output-format', dest='output_format', type=str,
        choices=['text', 'json'], action='store', default='text',
        metavar='FORMAT', help=(
            'Format of the output, the default is: text. Supported options: '
            'json, text.'))

    argument_parser.add_argument(
        '-v', '--verbose', dest='verbose', action='store_true',
        default=False, help='Print verbose output.')

    argument_parser.add_argument(
        '-w', '--write', metavar='OUTPUTFILE', dest='write',
        help='Output filename.')

    try:
      options = argument_parser.parse_args(arguments)
    except UnicodeEncodeError:
      # If we get here we are attempting to print help in a non-Unicode
      # terminal.
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_help())
      return False

    try:
      self.ParseOptions(options)
    except errors.BadConfigOption as exception:
      self._output_writer.Write('ERROR: {0!s}\n'.format(exception))
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_usage())
      return False

    loggers.ConfigureLogging(
        debug_output=self._debug_mode, filename=self._log_file,
        quiet_mode=self._quiet_mode)

    return True

  def ParseOptions(self, options):
    """Parses the options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    self._ParseInformationalOptions(options)

    self._verbose = getattr(options, 'verbose', False)

    self.show_troubleshooting = getattr(options, 'show_troubleshooting', False)
    if self.show_troubleshooting:
      return

    self._output_filename = getattr(options, 'write', None)

    argument_helper_names = ['process_resources', 'storage_file']
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=argument_helper_names)

    # TODO: move check into _CheckStorageFile.
    if not self._storage_file_path:
      raise errors.BadConfigOption('Missing storage file option.')

    if not os.path.isfile(self._storage_file_path):
      raise errors.BadConfigOption(
          'No such storage file: {0:s}.'.format(self._storage_file_path))

    compare_storage_file_path = self.ParseStringOption(
        options, 'compare_storage_file')
    if compare_storage_file_path:
      if not os.path.isfile(compare_storage_file_path):
        raise errors.BadConfigOption(
            'No such storage file: {0:s}.'.format(compare_storage_file_path))

      self._compare_storage_file_path = compare_storage_file_path
      self.compare_storage_information = True

    self._output_format = self.ParseStringOption(options, 'output_format')

    if self._output_filename:
      if os.path.exists(self._output_filename):
        raise errors.BadConfigOption(
            'Output file already exists: {0:s}.'.format(self._output_filename))
      output_file_object = open(self._output_filename, 'wb')
      self._output_writer = tools.FileObjectOutputWriter(output_file_object)

    self._EnforceProcessMemoryLimit(self._process_memory_limit)

  def PrintStorageInformation(self):
    """Prints the storage information."""
    storage_reader = storage_factory.StorageFactory.CreateStorageReaderForFile(
        self._storage_file_path)
    if not storage_reader:
      logger.error(
          'Format of storage file: {0:s} not supported'.format(
              self._storage_file_path))
      return

    try:
      if self._output_format == 'json':
        self._PrintStorageInformationAsJSON(storage_reader)
      elif self._output_format == 'text':
        self._PrintStorageInformationAsText(storage_reader)
    finally:
      storage_reader.Close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the SHA-1 hasher implementation."""

from __future__ import unicode_literals

import unittest

from plaso.analyzers.hashers import sha1

from tests.analyzers.hashers import test_lib


class SHA1Test(test_lib.HasherTestCase):
  """Tests the SHA-1 hasher."""

  def testFileHashMatchesEmptyFile(self):
    """Tests that hasher matches the hash of an empty file."""
    hasher = sha1.SHA1Hasher()
    self._AssertTestPathStringDigestMatch(
        hasher, ['empty_file'], 'da39a3ee5e6b4b0d3255bfef95601890afd80709')

  def testFileHashMatchesKnownFile(self):
    """Tests that hasher matches the hash of an empty file."""
    hasher = sha1.SHA1Hasher()
    self._AssertTestPathStringDigestMatch(
        hasher, ['ímynd.dd'], 'd9f264323004fd9518c0474967f80421e60e9813')


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The Kodi MyVideos database event formatter."""

from __future__ import unicode_literals

from plaso.formatters import interface
from plaso.formatters import manager


class KodiFormatter(interface.ConditionalEventFormatter):
  """Formatter for an Kodi Video event."""

  DATA_TYPE = 'kodi:videos:viewing'

  FORMAT_STRING_PIECES = [
      'Video: {filename}',
      'Play Count: {play_count}']

  FORMAT_STRING_SHORT_PIECES = ['{filename}']

  SOURCE_LONG = 'Kodi Video Viewed'
  SOURCE_SHORT = 'KODI'


manager.FormattersManager.RegisterFormatter(KodiFormatter)



# -*- coding: utf-8 -*-
"""A plugin to enable quick triage of Windows Services."""

from __future__ import unicode_literals

import yaml

from plaso.analysis import interface
from plaso.analysis import manager
from plaso.containers import reports
from plaso.lib import py2to3
from plaso.winnt import human_readable_service_enums


class WindowsService(yaml.YAMLObject):
  """Class to represent a Windows Service.

  Attributes:
    image_path (str): value of the ImagePath value of the service key.
    name (str): name of the service
    object_name (str): value of the ObjectName value of the service key.
    service_dll (str): value of the ServiceDll value in the service's Parameters
        subkey.
    service_type (int): value of the Type value of the service key.
    source (tuple[str, str]) : tuple containing the path and registry key
        describing where the service was found
    start_type (int):  value of the Start value of the service key.
    """
  # This is used for comparison operations and defines attributes that should
  # not be used during evaluation of whether two services are the same.
  COMPARE_EXCLUDE = frozenset(['sources'])

  _REGISTRY_KEY_PATH_SEPARATOR = '\\'

  # YAML attributes
  yaml_tag = '!WindowsService'
  yaml_loader = yaml.SafeLoader
  yaml_dumper = yaml.SafeDumper

  def __init__(
      self, name, service_type, image_path, start_type, object_name, source,
      service_dll=None):
    """Initializes a Windows service object.

    Args:
      name (str): name of the service
      service_type (int): value of the Type value of the service key.
      image_path (str): value of the ImagePath value of the service key.
      start_type (int):  value of the Start value of the service key.
      object_name (str): value of the ObjectName value of the service key.
      source (tuple[str, str]) : tuple containing the path and registry key
          describing where the service was found
      service_dll (Optional[str]): value of the ServiceDll value in the
          service's Parameters subkey.

    Raises:
      TypeError: If a tuple with two elements is not passed as the 'source'
                 argument.
    """
    super(WindowsService, self).__init__()
    self.anomalies = []
    self.image_path = image_path
    self.name = name
    self.object_name = object_name
    self.service_dll = service_dll
    self.service_type = service_type
    self.start_type = start_type

    if isinstance(source, tuple):
      if len(source) != 2:
        raise TypeError('Source arguments must be tuple of length 2.')
      # A service may be found in multiple Control Sets or Registry hives,
      # hence the list.
      self.sources = [source]
    else:
      raise TypeError('Source argument must be a tuple.')

  def __eq__(self, other_service):
    """Custom equality method so that we match near-duplicates.

    Compares two service objects together and evaluates if they are
    the same or close enough to be considered to represent the same service.

    For two service objects to be considered the same they need to
    have the the same set of attributes and same values for all their
    attributes, other than those enumerated as reserved in the
    COMPARE_EXCLUDE constant.

    Args:
      other_service (WindowsService): service we are testing for equality.

    Returns:
      bool: whether the services are equal.
    """
    if not isinstance(other_service, WindowsService):
      return False

    attributes = set(self.__dict__.keys())
    other_attributes = set(self.__dict__.keys())

    if attributes != other_attributes:
      return False

    # We compare the values for all attributes, other than those specifically
    # enumerated as not relevant for equality comparisons.
    for attribute in attributes.difference(self.COMPARE_EXCLUDE):
      if getattr(self, attribute, None) != getattr(
          other_service, attribute, None):
        return False

    return True

  @classmethod
  def FromEventData(cls, event_data):
    """Creates a service object from event data.

    Args:
      event_data (EventData): event data.

    Returns:
      WindowsService: service.
    """
    if event_data.pathspec:
      source = (event_data.pathspec.location, event_data.key_path)
    else:
      source = ('Unknown', 'Unknown')

    return cls(
        event_data.name, event_data.service_type, event_data.image_path,
        event_data.start_type, event_data.object_name, source,
        service_dll=event_data.service_dll)

  def HumanReadableType(self):
    """Return a human readable string describing the type value.

    Returns:
      str: human readable description of the type value.
    """
    if isinstance(self.service_type, py2to3.STRING_TYPES):
      return self.service_type
    return human_readable_service_enums.SERVICE_ENUMS['Type'].get(
        self.service_type, '{0:d}'.format(self.service_type))

  def HumanReadableStartType(self):
    """Return a human readable string describing the start type value.

    Returns:
      str: human readable description of the start type value.
    """
    if isinstance(self.start_type, py2to3.STRING_TYPES):
      return self.start_type
    return human_readable_service_enums.SERVICE_ENUMS['Start'].get(
        self.start_type, '{0:d}'.format(self.start_type))


class WindowsServiceCollection(object):
  """Class to hold and de-duplicate Windows Services."""

  def __init__(self):
    """Initialize a collection that holds Windows Service."""
    super(WindowsServiceCollection, self).__init__()
    self._services = []

  def AddService(self, new_service):
    """Add a new service to the list of ones we know about.

    Args:
      new_service (WindowsService): the service to add.
    """
    for service in self._services:
      if new_service == service:
        # If this service is the same as one we already know about, we
        # just want to add where it came from.
        service.sources.append(new_service.sources[0])
        return

    # We only add a new object to our list if we don't have
    # an identical one already.
    self._services.append(new_service)

  @property
  def services(self):
    """list[WindowsService]: services in this collection."""
    return self._services


class WindowsServicesAnalysisPlugin(interface.AnalysisPlugin):
  """Provides a single list of for Windows services found in the Registry."""

  NAME = 'windows_services'

  # Indicate that we can run this plugin during regular extraction.
  ENABLE_IN_EXTRACTION = True

  _SUPPORTED_EVENT_DATA_TYPES = frozenset([
      'windows:registry:service'])

  def __init__(self):
    """Initializes the Windows Services plugin."""
    super(WindowsServicesAnalysisPlugin, self).__init__()
    self._output_format = 'text'
    self._service_collection = WindowsServiceCollection()

  def _FormatServiceText(self, service):
    """Produces a human readable multi-line string representing the service.

    Args:
      service (WindowsService):  service to format.

    Returns:
      str: human readable representation of a Windows Service.
    """
    service_type = service.HumanReadableType()
    start_type = service.HumanReadableStartType()

    string_segments = [
        service.name,
        '\tImage Path    = {0:s}'.format(service.image_path or ''),
        '\tService Type  = {0:s}'.format(service_type),
        '\tStart Type    = {0:s}'.format(start_type),
        '\tService Dll   = {0:s}'.format(service.service_dll or ''),
        '\tObject Name   = {0:s}'.format(service.object_name or ''),
        '\tSources:']

    string_segments.extend([
        '\t\t{0:s}:{1:s}'.format(source[0], source[1])
        for source in service.sources])
    return '\n'.join(string_segments)

  def CompileReport(self, mediator):
    """Compiles an analysis report.

    Args:
      mediator (AnalysisMediator): mediates interactions between analysis
          plugins and other components, such as storage and dfvfs.

    Returns:
      AnalysisReport: report.
    """
    # TODO: move YAML representation out of plugin and into serialization.
    lines_of_text = []
    if self._output_format == 'yaml':
      lines_of_text.append(
          yaml.safe_dump_all(self._service_collection.services))
    else:
      lines_of_text.append('Listing Windows Services')
      for service in self._service_collection.services:
        lines_of_text.append(self._FormatServiceText(service))
        lines_of_text.append('')

    lines_of_text.append('')
    report_text = '\n'.join(lines_of_text)
    return reports.AnalysisReport(plugin_name=self.NAME, text=report_text)

  # pylint: disable=unused-argument
  def ExamineEvent(self, mediator, event, event_data):
    """Analyzes an event and creates Windows Services as required.

    At present, this method only handles events extracted from the Registry.

    Args:
      mediator (AnalysisMediator): mediates interactions between analysis
          plugins and other components, such as storage and dfvfs.
      event (EventObject): event to examine.
      event_data (EventData): event data.
    """
    if event_data.data_type not in self._SUPPORTED_EVENT_DATA_TYPES:
      return

    # TODO: Handle event log entries here also (ie, event id 4697).
    service = WindowsService.FromEventData(event_data)
    self._service_collection.AddService(service)

  def SetOutputFormat(self, output_format):
    """Sets the output format of the generated report.

    Args:
      output_format (str): format the plugin should used to produce its output.
    """
    self._output_format = output_format


manager.AnalysisPluginManager.RegisterPlugin(WindowsServicesAnalysisPlugin)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the sessionize analysis plugin CLI arguments helper."""

from __future__ import unicode_literals

import argparse
import unittest

from plaso.analysis import sessionize
from plaso.cli.helpers import sessionize_analysis
from plaso.lib import errors

from tests.cli import test_lib as cli_test_lib
from tests.cli.helpers import test_lib


class SessionizeAnalysisArgumentsHelperTest(
    test_lib.AnalysisPluginArgumentsHelperTest):
  """Tests the sessionize analysis plugin CLI arguments helper."""

  # pylint: disable=no-member,protected-access

  _EXPECTED_OUTPUT = """\
usage: cli_helper.py [--maximum-pause MINUTES]

Test argument parser.

optional arguments:
  --maximum-pause MINUTES, --maximum_pause MINUTES
                        Specify the maximum delay in minutes between events in
                        the session.
"""

  def testAddArguments(self):
    """Tests the AddArguments function."""
    argument_parser = argparse.ArgumentParser(
        prog='cli_helper.py',
        description='Test argument parser.', add_help=False,
        formatter_class=cli_test_lib.SortedArgumentsHelpFormatter)

    sessionize_analysis.SessionizeAnalysisArgumentsHelper.AddArguments(
        argument_parser)

    output = self._RunArgparseFormatHelp(argument_parser)
    self.assertEqual(output, self._EXPECTED_OUTPUT)

  def testParseOptions(self):
    """Tests the ParseOptions function."""
    options = cli_test_lib.TestOptions()

    analysis_plugin = sessionize.SessionizeAnalysisPlugin()
    sessionize_analysis.SessionizeAnalysisArgumentsHelper.ParseOptions(
        options, analysis_plugin)

    with self.assertRaises(errors.BadConfigObject):
      sessionize_analysis.SessionizeAnalysisArgumentsHelper.ParseOptions(
          options, None)

    options.sessionize_maximumpause = 0
    with self.assertRaises(errors.BadConfigOption):
      sessionize_analysis.SessionizeAnalysisArgumentsHelper.ParseOptions(
          options, analysis_plugin)

    options.sessionize_maximumpause = 'ten'
    with self.assertRaises(errors.BadConfigOption):
      sessionize_analysis.SessionizeAnalysisArgumentsHelper.ParseOptions(
          options, analysis_plugin)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""MountPoints2 Windows Registry parser plugin."""

from __future__ import unicode_literals

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import definitions
from plaso.parsers import winreg
from plaso.parsers.winreg_plugins import interface


class MountPoints2EventData(events.EventData):
  """Windows MountPoints2 event data attribute container.

  Attributes:
    key_path (str): Windows Registry key path.
    label (str): mount point label.
    name (str): name of the mount point source.
    server_name (str): name of the remote drive server or None if not set.
    share_name (str): name of the remote drive share or None if not set.
    type (str): type of the mount point source, which can be "Drive",
        "Remove Drive" or "Volume".
  """

  DATA_TYPE = 'windows:registry:mount_points2'

  def __init__(self):
    """Initializes event data."""
    super(MountPoints2EventData, self).__init__(data_type=self.DATA_TYPE)
    self.key_path = None
    self.label = None
    self.name = None
    self.server_name = None
    self.share_name = None
    self.type = None


class MountPoints2Plugin(interface.WindowsRegistryPlugin):
  """Windows Registry plugin for parsing the MountPoints2 key."""

  NAME = 'explorer_mountpoints2'
  DESCRIPTION = 'Parser for mount points Registry data.'

  FILTERS = frozenset([
      interface.WindowsRegistryKeyPathFilter(
          'HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\'
          'Explorer\\MountPoints2')])

  def ExtractEvents(self, parser_mediator, registry_key, **kwargs):
    """Extracts events from a Windows Registry key.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      registry_key (dfwinreg.WinRegistryKey): Windows Registry key.
    """
    for subkey in registry_key.GetSubkeys():
      name = subkey.name
      if not name:
        continue

      server_name = 'Drive'
      share_name = None
      source_type = None

      if name.startswith('{'):
        source_type = 'Volume'

      # Check if the name is formatted as: "##Server_Name#Share_Name".
      elif name.startswith('##'):
        source_type = 'Remote Drive'
        server_name, _, share_name = name[2:].partition('#')
        share_name = '\\{0:s}'.format(share_name.replace('#', '\\'))

      label_value = subkey.GetValueByName('_LabelFromReg')
      if label_value:
        label = label_value.GetDataAsObject()
      else:
        label = None

      event_data = MountPoints2EventData()
      event_data.key_path = registry_key.path
      event_data.label = label
      event_data.name = name
      event_data.server_name = server_name
      event_data.share_name = share_name
      event_data.type = source_type

      event = time_events.DateTimeValuesEvent(
          subkey.last_written_time, definitions.TIME_DESCRIPTION_WRITTEN)
      parser_mediator.ProduceEventWithEventData(event, event_data)


winreg.WinRegistryParser.RegisterPlugin(MountPoints2Plugin)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Microsoft Internet Explorer (MSIE) Cache Files (CF) parser."""

from __future__ import unicode_literals

import unittest

from plaso.formatters import msiecf as _  # pylint: disable=unused-import
from plaso.lib import definitions
from plaso.parsers import msiecf

from tests.parsers import test_lib


class MSIECFParserTest(test_lib.ParserTestCase):
  """Tests for the MSIE Cache Files (MSIECF) parser."""

  def testParse(self):
    """Tests the Parse function."""
    parser = msiecf.MSIECFParser()
    storage_writer = self._ParseFile(['index.dat'], parser)

    # MSIE Cache File information:
    #   Version                         : 5.2
    #   File size                       : 32768 bytes
    #   Number of items                 : 7
    #   Number of recovered items       : 11

    self.assertEqual(storage_writer.number_of_warnings, 0)
    # 7 + 11 records, each with 4 records.
    self.assertEqual(storage_writer.number_of_events, (7 + 11) * 4)

    events = list(storage_writer.GetEvents())

    # Record type             : URL
    # Offset range            : 21376 - 21632 (256)
    # Location                : Visited: testing@http://www.trafficfusionx.com
    #                           /download/tfscrn2/funnycats.exe
    # Primary time            : Jun 23, 2011 18:02:10.066000000
    # Secondary time          : Jun 23, 2011 18:02:10.066000000
    # Expiration time         : Jun 29, 2011 17:55:02
    # Last checked time       : Jun 23, 2011 18:02:12
    # Cache directory index   : -2 (0xfe)

    event = events[8]
    expected_url = (
        'Visited: testing@http://www.trafficfusionx.com/download/tfscrn2'
        '/funnycats.exe')

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.data_type, 'msiecf:url')
    self.assertEqual(event_data.offset, 21376)
    self.assertEqual(event_data.url, expected_url)
    self.assertEqual(event_data.cache_directory_index, -2)

    self.CheckTimestamp(event.timestamp, '2011-06-23 18:02:10.066000')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_LAST_VISITED)

    event = events[9]

    self.CheckTimestamp(event.timestamp, '2011-06-23 18:02:10.066000')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_LAST_VISITED)

    event = events[10]

    self.CheckTimestamp(event.timestamp, '2011-06-29 17:55:02.000000')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_EXPIRATION)

    event = events[11]

    self.CheckTimestamp(event.timestamp, '2011-06-23 18:02:12.000000')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_LAST_CHECKED)

    expected_message = (
        'Location: Visited: testing@http://www.trafficfusionx.com/download'
        '/tfscrn2/funnycats.exe '
        'Number of hits: 6 '
        'Cached file size: 0')
    expected_short_message = (
        'Location: Visited: testing@http://www.trafficfusionx.com/download'
        '/tfscrn2/fun...')

    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

  def testParseLeakAndRedirect(self):
    """Tests the Parse function with leak and redirected records."""
    parser = msiecf.MSIECFParser()
    storage_writer = self._ParseFile(['nfury_index.dat'], parser)
    self.assertEqual(storage_writer.number_of_warnings, 0)

    # MSIE Cache File information:
    #   Version                         : 5.2
    #   File size                       : 491520 bytes
    #   Number of items                 : 1027
    #   Number of recovered items       : 8

    self.assertEqual(storage_writer.number_of_events, 2898)

    events = list(storage_writer.GetEvents())

    event = events[3]

    event_data = self._GetEventDataOfEvent(storage_writer, event)

    # Test cached file path.
    self.assertEqual(event_data.data_type, 'msiecf:url')

    expected_message = (
        'Location: http://col.stc.s-msn.com/br/gbl/lg/csl/favicon.ico '
        'Number of hits: 1 '
        'Cached file: R6QWCVX4\\favicon[1].ico '
        'Cached file size: 4286 '
        'HTTP headers: HTTP/1.1 200 OK - '
        'Content-Type: image/x-icon - '
        'ETag: "0922651f38cb1:0", - '
        'X-Powered-By: ASP.NET - P3P: '
        'CP="BUS CUR CONo FIN IVDo ONL OUR PHY SAMo TELo" - '
        'Content-Length: 4286 - '
        ' - ~U:nfury - ')
    expected_short_message = (
        'Location: http://col.stc.s-msn.com/br/gbl/lg/csl/favicon.ico '
        'Cached file: R6Q...')

    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

    event = events[21]
    expected_url = (
        'http://ad.doubleclick.net/ad/N2724.Meebo/B5343067.13;sz=1x1;'
        'pc=[TPAS_ID];ord=2642102')

    event = events[16]

    self.assertEqual(event.timestamp, 0)
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_NOT_A_TIME)

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.data_type, 'msiecf:leak')
    self.assertEqual(event_data.cache_directory_index, 1)
    self.assertEqual(event_data.cache_directory_name, 'VUQHQA73')
    self.assertEqual(event_data.cached_file_size, 1966)
    self.assertEqual(event_data.cached_filename, 'ADSAdClient31[1].htm')
    self.assertEqual(event_data.recovered, False)

    expected_message = (
        'Cached file: VUQHQA73\\ADSAdClient31[1].htm '
        'Cached file size: 1966')
    expected_short_message = (
        'Cached file: VUQHQA73\\ADSAdClient31[1].htm')

    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

    event = events[21]
    self.assertEqual(event.timestamp, 0)
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_NOT_A_TIME)

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.data_type, 'msiecf:redirected')
    expected_url = (
        'http://ad.doubleclick.net/ad/N2724.Meebo/B5343067.13;sz=1x1;'
        'pc=[TPAS_ID];ord=2642102')
    self.assertEqual(event_data.url, expected_url)
    self.assertEqual(event_data.recovered, False)

    expected_message = 'Location: {0:s}'.format(expected_url)
    expected_short_message = (
        'Location: http://ad.doubleclick.net/ad/N2724.Meebo/B5343067.13;'
        'sz=1x1;pc=[TPA...')

    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Parser for the MacOS Duet/KnowledgeC database."""

from __future__ import unicode_literals

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import definitions
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class MacKnowledgeCApplicationEventData(events.EventData):
  """KnowledgeC application execution event data.

  Attributes:
    bundle_identifier (str): bundle identifier of the application.
    duration (int): duration of the activity.
  """

  DATA_TYPE = 'mac:knowledgec:application'

  def __init__(self):
    """Initializes event data."""
    super(MacKnowledgeCApplicationEventData, self).__init__(
        data_type=self.DATA_TYPE)
    self.bundle_identifier = None
    self.duration = None


class MacKnowledgeCSafariEventData(events.EventData):
  """MacOS Duet / KnowledgeC database event data for Safari.

  Attributes:
    bundle_identifier (str): bundle identifier of the application.
    duration (int): duration of the activity.
    title (str): title of the webpage visited.
    url (str): URL visited.
  """

  DATA_TYPE = 'mac:knowledgec:safari'

  def __init__(self):
    """Initializes event data."""
    super(MacKnowledgeCSafariEventData, self).__init__(data_type=self.DATA_TYPE)
    self.bundle_identifier = None
    self.duration = None
    self.title = None
    self.url = None


class MacKnowledgeCPlugin(interface.SQLitePlugin):
  """Parse the MacOS Duet / KnowledgeC database."""

  NAME = 'mac_knowledgec'
  DESCRIPTION = 'Parser for Duet / KnowledgeC database file.'

  # Define the needed queries.
  # entry_creation: when the entry was created in the database.
  # start: when the activity started.
  # end: when the activity finished.
  # action: action that created the entry (inFocus, activity, intents).
  # bundle_identifier: application bundle identifier.
  # title: Safari url title

  QUERIES = [
      (("""
        SELECT
          ZOBJECT.ZCREATIONDATE AS "entry_creation", 
          ZOBJECT.ZSTARTDATE AS "start", 
          ZOBJECT.ZENDDATE AS "end",
          ZOBJECT.ZSTREAMNAME AS "action",
          ZOBJECT.ZVALUESTRING AS "zvaluestring",
          ZSTRUCTUREDMETADATA.Z_DKSAFARIHISTORYMETADATAKEY__TITLE AS "title"
        FROM ZOBJECT
        LEFT JOIN ZSTRUCTUREDMETADATA 
        ON ZOBJECT.ZSTRUCTUREDMETADATA = ZSTRUCTUREDMETADATA.Z_PK
        """),
       'KnowledgeCRow')]

  # The required tables for the query.
  REQUIRED_TABLES = frozenset(['ZOBJECT', 'ZSTRUCTUREDMETADATA'])

  _SCHEMA_10_13 = {
      'ACHANGE': (
          'CREATE TABLE ACHANGE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZCHANGETYPE INTEGER, ZENTITY INTEGER, ZENTITYPK '
          'INTEGER, ZTRANSACTIONID INTEGER, ZCOLUMNS BLOB, ZTOMBSTONE0 BLOB, '
          'ZTOMBSTONE1 BLOB, ZTOMBSTONE2 BLOB )'),
      'ATRANSACTION': (
          'CREATE TABLE ATRANSACTION ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZTIMESTAMP FLOAT, ZAUTHOR VARCHAR, '
          'ZBUNDLEID VARCHAR, ZCONTEXTNAME VARCHAR, ZPROCESSID VARCHAR, '
          'ZQUERYGEN BLOB )'),
      'ZADDITIONCHANGESET': (
          'CREATE TABLE ZADDITIONCHANGESET ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZSEQUENCENUMBER INTEGER, ZVERSION INTEGER, '
          'ZENDDATE TIMESTAMP, ZSTARTDATE TIMESTAMP, ZCKFOREIGNKEY VARCHAR, '
          'ZCKRECORDID VARCHAR, ZDEVICEIDENTIFIER VARCHAR, ZCHANGESET BLOB, '
          'ZCKRECORDSYSTEMFIELDS BLOB )'),
      'ZCONTEXTUALCHANGEREGISTRATION': (
          'CREATE TABLE ZCONTEXTUALCHANGEREGISTRATION ( Z_PK INTEGER PRIMARY '
          'KEY, Z_ENT INTEGER, Z_OPT INTEGER, ZCREATIONDATE TIMESTAMP, '
          'ZIDENTIFIER VARCHAR, ZPROPERTIES BLOB )'),
      'ZCONTEXTUALKEYPATH': (
          'CREATE TABLE ZCONTEXTUALKEYPATH ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZDEVICEID INTEGER, ZISEPHEMERAL INTEGER, '
          'ZISUSERCENTRIC INTEGER, ZCREATIONDATE TIMESTAMP, ZLASTMODIFIEDDATE '
          'TIMESTAMP, ZKEY VARCHAR, ZVALUE BLOB )'),
      'ZCUSTOMMETADATA': (
          'CREATE TABLE ZCUSTOMMETADATA ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZINTEGERVALUE INTEGER, ZOBJECT INTEGER, '
          'Z8_OBJECT INTEGER, ZDATEVALUE TIMESTAMP, ZDOUBLEVALUE FLOAT, ZNAME '
          'VARCHAR, ZSTRINGVALUE VARCHAR, ZVALUEHASH VARCHAR, ZBINARYVALUE '
          'BLOB )'),
      'ZDELETIONCHANGESET': (
          'CREATE TABLE ZDELETIONCHANGESET ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZSEQUENCENUMBER INTEGER, ZVERSION INTEGER, '
          'ZENDDATE TIMESTAMP, ZSTARTDATE TIMESTAMP, ZCKFOREIGNKEY VARCHAR, '
          'ZCKRECORDID VARCHAR, ZDEVICEIDENTIFIER VARCHAR, ZCHANGESET BLOB, '
          'ZCKRECORDSYSTEMFIELDS BLOB )'),
      'ZHISTOGRAM': (
          'CREATE TABLE ZHISTOGRAM ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZSTREAMTYPECODE INTEGER, ZENDDATE TIMESTAMP, '
          'ZSTARTDATE TIMESTAMP, ZDEVICEIDENTIFIER VARCHAR, ZIDENTIFIER '
          'VARCHAR, ZSTREAMNAME VARCHAR )'),
      'ZHISTOGRAMVALUE': (
          'CREATE TABLE ZHISTOGRAMVALUE ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZCOUNT INTEGER, ZINTEGERVALUE INTEGER, '
          'ZHISTOGRAM INTEGER, ZSTRINGVALUE VARCHAR )'),
      'ZOBJECT': (
          'CREATE TABLE ZOBJECT ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZUUIDHASH INTEGER, ZEVENT INTEGER, ZSOURCE INTEGER, '
          'ZCATEGORYTYPE INTEGER, ZINTEGERVALUE INTEGER, ZENDDAYOFWEEK '
          'INTEGER, ZENDSECONDOFDAY INTEGER, ZHASCUSTOMMETADATA INTEGER, '
          'ZHASSTRUCTUREDMETADATA INTEGER, ZSECONDSFROMGMT INTEGER, '
          'ZSHOULDSYNC INTEGER, ZSTARTDAYOFWEEK INTEGER, ZSTARTSECONDOFDAY '
          'INTEGER, ZVALUECLASS INTEGER, ZVALUEINTEGER INTEGER, '
          'ZVALUETYPECODE INTEGER, ZSTRUCTUREDMETADATA INTEGER, ZVALUE '
          'INTEGER, Z8_VALUE INTEGER, ZIDENTIFIERTYPE INTEGER, ZQUANTITYTYPE '
          'INTEGER, ZOBJECT INTEGER, Z8_OBJECT INTEGER, ZSUBJECT INTEGER, '
          'Z8_SUBJECT INTEGER, ZCREATIONDATE TIMESTAMP, ZCONFIDENCE FLOAT, '
          'ZENDDATE TIMESTAMP, ZSTARTDATE TIMESTAMP, ZVALUEDOUBLE FLOAT, '
          'ZDOUBLEVALUE FLOAT, ZUUID VARCHAR, ZSTREAMNAME VARCHAR, '
          'ZVALUESTRING VARCHAR, ZSTRING VARCHAR, ZVERBPHRASE VARCHAR, '
          'ZMETADATA BLOB )'),
      'ZSOURCE': (
          'CREATE TABLE ZSOURCE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZUSERID INTEGER, ZBUNDLEID VARCHAR, ZDEVICEID '
          'VARCHAR, ZGROUPID VARCHAR, ZITEMID VARCHAR, ZSOURCEID VARCHAR )'),
      'ZSTRUCTUREDMETADATA': (
          'CREATE TABLE ZSTRUCTUREDMETADATA ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, Z_CDPORTRAITMETADATAKEY__ALGORITHM '
          'INTEGER, Z_CDPORTRAITMETADATAKEY__ASSETVERSION INTEGER, '
          'Z_DKAPPINSTALLMETADATAKEY__ISINSTALL INTEGER, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ISPUBLICLYINDEXABLE INTEGER, '
          'Z_DKAPPLICATIONMETADATAKEY__PROCESSIDENTIFIER INTEGER, '
          'Z_DKAUDIOMETADATAKEY__ROUTECHANGEREASON INTEGER, '
          'Z_DKBLUETOOTHMETADATAKEY__DEVICETYPE INTEGER, '
          'Z_DKBULLETINBOARDMETADATAKEY__HASDATE INTEGER, '
          'Z_DKGLANCELAUNCHMETADATA__DEVICEIDENTIFIER INTEGER, '
          'Z_DKINTENTMETADATAKEY__DONATEDBYSIRI INTEGER, '
          'Z_DKINTENTMETADATAKEY__INTENTHANDLINGSTATUS INTEGER, '
          'Z_DKNOWPLAYINGMETADATAKEY__IDENTIFIER INTEGER, '
          'Z_DKNOWPLAYINGMETADATAKEY__PLAYING INTEGER, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__INTERACTIONTYPE INTEGER, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__SUGGESTIONTYPE INTEGER, '
          'Z_DKSUNRISESUNSETMETADATAKEY__ISDAYLIGHT INTEGER, '
          'Z_QPMETRICSMETADATAKEY__QUERYENGAGED INTEGER, '
          'Z_QPMETRICSMETADATAKEY__RESULTENGAGED INTEGER, '
          'ZCOM_APPLE_CALENDARUIKIT_USERACTIVITY_DATE INTEGER, '
          'ZCOM_APPLE_CALENDARUIKIT_USERACTIVITY_ENDDATE INTEGER, '
          'Z_CDPORTRAITMETADATAKEY__DECAYRATE FLOAT, '
          'Z_CDPORTRAITMETADATAKEY__SCORE FLOAT, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__EXPIRATIONDATE TIMESTAMP, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__LATITUDE FLOAT, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__LONGITUDE FLOAT, '
          'Z_DKLOCATIONMETADATAKEY__LATITUDE FLOAT, '
          'Z_DKLOCATIONMETADATAKEY__LONGITUDE FLOAT, '
          'Z_DKNOWPLAYINGMETADATAKEY__DURATION FLOAT, '
          'Z_DKNOWPLAYINGMETADATAKEY__ELAPSED FLOAT, '
          'Z_DKPERIODMETADATAKEY__PERIODEND TIMESTAMP, '
          'Z_DKPERIODMETADATAKEY__PERIODSTART TIMESTAMP, '
          'Z_DKSUNRISESUNSETMETADATAKEY__CURRENTSUNRISE TIMESTAMP, '
          'Z_DKSUNRISESUNSETMETADATAKEY__CURRENTSUNSET TIMESTAMP, '
          'Z_DKSUNRISESUNSETMETADATAKEY__NEXTSUNRISE TIMESTAMP, '
          'Z_DKSUNRISESUNSETMETADATAKEY__NEXTSUNSET TIMESTAMP, '
          'Z_DKSUNRISESUNSETMETADATAKEY__PREVIOUSSUNRISE TIMESTAMP, '
          'Z_DKSUNRISESUNSETMETADATAKEY__PREVIOUSSUNSET TIMESTAMP, '
          'Z_QPMETRICSMETADATAKEY__TIMESTAMP FLOAT, '
          'Z_CDENTITYMETADATAKEY__BESTLANGUAGE VARCHAR, '
          'Z_CDENTITYMETADATAKEY__NAME VARCHAR, '
          'Z_CDPORTRAITMETADATAKEY__OSBUILD VARCHAR, '
          'Z_DKAPPINSTALLMETADATAKEY__PRIMARYCATEGORY VARCHAR, '
          'Z_DKAPPINSTALLMETADATAKEY__TITLE VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ACTIVITYTYPE VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ITEMIDENTIFIER VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ITEMRELATEDUNIQUEIDENTIFIER '
          'VARCHAR, Z_DKAPPLICATIONACTIVITYMETADATAKEY__TITLE VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__USERACTIVITYREQUIREDSTRING '
          'VARCHAR, Z_DKAPPLICATIONACTIVITYMETADATAKEY__USERACTIVITYUUID '
          'VARCHAR, Z_DKAPPLICATIONMETADATAKEY__BACKBOARDSTATE VARCHAR, '
          'Z_DKAPPLICATIONMETADATAKEY__EXTENSIONCONTAININGBUNDLEIDENTIFIER '
          'VARCHAR, Z_DKAPPLICATIONMETADATAKEY__EXTENSIONHOSTIDENTIFIER '
          'VARCHAR, Z_DKAPPLICATIONMETADATAKEY__LAUNCHREASON VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__CHANNELS VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__DATASOURCES VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__IDENTIFIER VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__PORTNAME VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__PORTTYPE VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__PREFERREDDATASOURCE VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__SELECTEDDATASOURCE VARCHAR, '
          'Z_DKBATTERYSAVERMETADATAKEY__SOURCE VARCHAR, '
          'Z_DKBLUETOOTHMETADATAKEY__ADDRESS VARCHAR, '
          'Z_DKBLUETOOTHMETADATAKEY__NAME VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__FEED VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__MESSAGE VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__SUBTITLE VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__TITLE VARCHAR, '
          'Z_DKCALENDARMETADATAKEY__INTERACTION VARCHAR, '
          'Z_DKCALLMETADATAKEY__INTERACTION VARCHAR, '
          'Z_DKDEVICEIDMETADATAKEY__DEVICEIDENTIFIER VARCHAR, '
          'Z_DKINTENTMETADATAKEY__INTENTCLASS VARCHAR, '
          'Z_DKINTENTMETADATAKEY__INTENTVERB VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__URL VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__CITY VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__COUNTRY VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__DISPLAYNAME VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__FULLYFORMATTEDADDRESS '
          'VARCHAR, Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__LOCATIONNAME '
          'VARCHAR, Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__POSTALCODE_V2 '
          'VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__STATEORPROVINCE '
          'VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__SUBTHOROUGHFARE '
          'VARCHAR, Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__THOROUGHFARE '
          'VARCHAR, Z_DKLOCATIONMETADATAKEY__IDENTIFIER VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__HOMEUUID VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__VIEWINFORMATION VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__VIEWNAME VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__VIEWUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__ACCESSORYNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__ACCESSORYUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__CHARACTERISTICTYPE VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__CLIENTNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__HOMEUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__SERVICENAME VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__SERVICETYPE VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__ACTIONSETNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__ACTIONSETTYPE VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__ACTIONSETUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__CLIENTNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__HOMEUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__SCENENAME VARCHAR, '
          'Z_DKMICROLOCATIONMETADATAKEY__LOCATIONDISTRIBUTION VARCHAR, '
          'Z_DKMICROLOCATIONMETADATAKEY__MICROLOCATIONDISTRIBUTION VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__ALBUM VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__ARTIST VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__GENRE VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__TITLE VARCHAR, '
          'Z_DKSAFARIHISTORYMETADATAKEY__TITLE VARCHAR, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__CLIENT VARCHAR, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__CONTACTID VARCHAR, '
          'Z_QPMETRICSMETADATAKEY__QUERY VARCHAR, '
          'ZCOM_APPLE_CALENDARUIKIT_USERACTIVITY_EXTERNALID VARCHAR, '
          'ZKCDCSNOTIFICATIONOPTIONCLIENTIDENTIFIERKEY VARCHAR, '
          'ZKCDCSNOTIFICATIONOPTIONCLIENTLAUNCHKEY VARCHAR, '
          'ZKCDCSNOTIFICATIONOPTIONPERSISTENTPREDICATESTRINGKEY VARCHAR, '
          'ZMETADATAHASH VARCHAR UNIQUE, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ITEMRELATEDCONTENTURL VARCHAR, '
          'Z_DKAPPINSTALLMETADATAKEY__SUBCATEGORIES BLOB, '
          'Z_DKINTENTMETADATAKEY__SERIALIZEDINTERACTION BLOB, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__PHONENUMBERS BLOB, '
          'Z_QPMETRICSMETADATAKEY__CANDIDATELIST BLOB, '
          'Z_QPMETRICSMETADATAKEY__QUERYLIST BLOB )'),
      'Z_4EVENT': (
          'CREATE TABLE Z_4EVENT ( Z_4CUSTOMMETADATA INTEGER, Z_10EVENT '
          'INTEGER, PRIMARY KEY (Z_4CUSTOMMETADATA, Z_10EVENT) )'),
      'Z_METADATA': (
          'CREATE TABLE Z_METADATA (Z_VERSION INTEGER PRIMARY KEY, Z_UUID '
          'VARCHAR(255), Z_PLIST BLOB)'),
      'Z_MODELCACHE': (
          'CREATE TABLE Z_MODELCACHE (Z_CONTENT BLOB)'),
      'Z_PRIMARYKEY': (
          'CREATE TABLE Z_PRIMARYKEY (Z_ENT INTEGER PRIMARY KEY, Z_NAME '
          'VARCHAR, Z_SUPER INTEGER, Z_MAX INTEGER)')}

  _SCHEMA_10_14 = {
      'ZADDITIONCHANGESET': (
          'CREATE TABLE ZADDITIONCHANGESET ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZSEQUENCENUMBER INTEGER, ZVERSION INTEGER, '
          'ZENDDATE TIMESTAMP, ZSTARTDATE TIMESTAMP, ZCKFOREIGNKEY VARCHAR, '
          'ZCKRECORDID VARCHAR, ZDEVICEIDENTIFIER VARCHAR, ZCHANGESET BLOB, '
          'ZCKRECORDSYSTEMFIELDS BLOB )'),
      'ZCONTEXTUALCHANGEREGISTRATION': (
          'CREATE TABLE ZCONTEXTUALCHANGEREGISTRATION ( Z_PK INTEGER PRIMARY '
          'KEY, Z_ENT INTEGER, Z_OPT INTEGER, ZCREATIONDATE TIMESTAMP, '
          'ZIDENTIFIER VARCHAR, ZPROPERTIES BLOB )'),
      'ZCONTEXTUALKEYPATH': (
          'CREATE TABLE ZCONTEXTUALKEYPATH ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZDEVICEID INTEGER, ZISEPHEMERAL INTEGER, '
          'ZISUSERCENTRIC INTEGER, ZCREATIONDATE TIMESTAMP, ZLASTMODIFIEDDATE '
          'TIMESTAMP, ZKEY VARCHAR, ZVALUE BLOB )'),
      'ZCUSTOMMETADATA': (
          'CREATE TABLE ZCUSTOMMETADATA ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZINTEGERVALUE INTEGER, ZOBJECT INTEGER, '
          'Z9_OBJECT INTEGER, ZDATEVALUE TIMESTAMP, ZDOUBLEVALUE FLOAT, ZNAME '
          'VARCHAR, ZSTRINGVALUE VARCHAR, ZVALUEHASH VARCHAR, ZBINARYVALUE '
          'BLOB )'),
      'ZDELETIONCHANGESET': (
          'CREATE TABLE ZDELETIONCHANGESET ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZSEQUENCENUMBER INTEGER, ZVERSION INTEGER, '
          'ZENDDATE TIMESTAMP, ZSTARTDATE TIMESTAMP, ZCKFOREIGNKEY VARCHAR, '
          'ZCKRECORDID VARCHAR, ZDEVICEIDENTIFIER VARCHAR, ZCHANGESET BLOB, '
          'ZCKRECORDSYSTEMFIELDS BLOB )'),
      'ZHISTOGRAM': (
          'CREATE TABLE ZHISTOGRAM ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZSTREAMTYPECODE INTEGER, ZENDDATE TIMESTAMP, '
          'ZSTARTDATE TIMESTAMP, ZCUSTOMIDENTIFIER VARCHAR, ZDEVICEIDENTIFIER '
          'VARCHAR, ZIDENTIFIER VARCHAR, ZSTREAMNAME VARCHAR )'),
      'ZHISTOGRAMVALUE': (
          'CREATE TABLE ZHISTOGRAMVALUE ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, ZINTEGERVALUE INTEGER, ZHISTOGRAM INTEGER, '
          'ZCOUNT FLOAT, ZSTRINGVALUE VARCHAR )'),
      'ZKEYVALUE': (
          'CREATE TABLE ZKEYVALUE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZDOMAIN VARCHAR, ZKEY VARCHAR, ZVALUE BLOB )'),
      'ZOBJECT': (
          'CREATE TABLE ZOBJECT ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZUUIDHASH INTEGER, ZEVENT INTEGER, ZSOURCE INTEGER, '
          'ZCATEGORYTYPE INTEGER, ZINTEGERVALUE INTEGER, ZENDDAYOFWEEK '
          'INTEGER, ZENDSECONDOFDAY INTEGER, ZHASCUSTOMMETADATA INTEGER, '
          'ZHASSTRUCTUREDMETADATA INTEGER, ZSECONDSFROMGMT INTEGER, '
          'ZSHOULDSYNC INTEGER, ZSTARTDAYOFWEEK INTEGER, ZSTARTSECONDOFDAY '
          'INTEGER, ZVALUECLASS INTEGER, ZVALUEINTEGER INTEGER, '
          'ZVALUETYPECODE INTEGER, ZSTRUCTUREDMETADATA INTEGER, ZVALUE '
          'INTEGER, Z9_VALUE INTEGER, ZIDENTIFIERTYPE INTEGER, ZQUANTITYTYPE '
          'INTEGER, ZOBJECT INTEGER, Z9_OBJECT INTEGER, ZSUBJECT INTEGER, '
          'Z9_SUBJECT INTEGER, ZCREATIONDATE TIMESTAMP, ZLOCALCREATIONDATE '
          'TIMESTAMP, ZCONFIDENCE FLOAT, ZENDDATE TIMESTAMP, ZSTARTDATE '
          'TIMESTAMP, ZVALUEDOUBLE FLOAT, ZDOUBLEVALUE FLOAT, ZUUID VARCHAR, '
          'ZSTREAMNAME VARCHAR, ZVALUESTRING VARCHAR, ZSTRING VARCHAR, '
          'ZVERBPHRASE VARCHAR, ZMETADATA BLOB )'),
      'ZSOURCE': (
          'CREATE TABLE ZSOURCE ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZUSERID INTEGER, ZBUNDLEID VARCHAR, ZDEVICEID '
          'VARCHAR, ZGROUPID VARCHAR, ZITEMID VARCHAR, ZSOURCEID VARCHAR )'),
      'ZSTRUCTUREDMETADATA': (
          'CREATE TABLE ZSTRUCTUREDMETADATA ( Z_PK INTEGER PRIMARY KEY, Z_ENT '
          'INTEGER, Z_OPT INTEGER, Z_CDPORTRAITMETADATAKEY__ALGORITHM '
          'INTEGER, Z_CDPORTRAITMETADATAKEY__ASSETVERSION INTEGER, '
          'Z_DKAPPINSTALLMETADATAKEY__ISINSTALL INTEGER, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ISELIGIBLEFORPREDICTION '
          'INTEGER, Z_DKAPPLICATIONACTIVITYMETADATAKEY__ISPUBLICLYINDEXABLE '
          'INTEGER, Z_DKAPPLICATIONMETADATAKEY__PROCESSIDENTIFIER INTEGER, '
          'Z_DKAUDIOMETADATAKEY__ROUTECHANGEREASON INTEGER, '
          'Z_DKBLUETOOTHMETADATAKEY__DEVICETYPE INTEGER, '
          'Z_DKBULLETINBOARDMETADATAKEY__HASDATE INTEGER, '
          'Z_DKDIGITALHEALTHMETADATAKEY__USAGETYPE INTEGER, '
          'Z_DKGLANCELAUNCHMETADATA__DEVICEIDENTIFIER INTEGER, '
          'Z_DKINTENTMETADATAKEY__DONATEDBYSIRI INTEGER, '
          'Z_DKINTENTMETADATAKEY__INTENTHANDLINGSTATUS INTEGER, '
          'Z_DKINTENTMETADATAKEY__INTENTTYPE INTEGER, '
          'Z_DKNOWPLAYINGMETADATAKEY__IDENTIFIER INTEGER, '
          'Z_DKNOWPLAYINGMETADATAKEY__PLAYING INTEGER, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__INTERACTIONTYPE INTEGER, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__SUGGESTIONTYPE INTEGER, '
          'Z_QPMETRICSMETADATAKEY__QUERYENGAGED INTEGER, '
          'Z_QPMETRICSMETADATAKEY__RESULTENGAGED INTEGER, '
          'ZCOM_APPLE_CALENDARUIKIT_USERACTIVITY_DATE INTEGER, '
          'ZCOM_APPLE_CALENDARUIKIT_USERACTIVITY_ENDDATE INTEGER, '
          'Z_CDPORTRAITMETADATAKEY__DECAYRATE FLOAT, '
          'Z_CDPORTRAITMETADATAKEY__SCORE FLOAT, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__EXPIRATIONDATE TIMESTAMP, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__LATITUDE FLOAT, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__LONGITUDE FLOAT, '
          'Z_DKLOCATIONMETADATAKEY__LATITUDE FLOAT, '
          'Z_DKLOCATIONMETADATAKEY__LONGITUDE FLOAT, '
          'Z_DKNOWPLAYINGMETADATAKEY__DURATION FLOAT, '
          'Z_DKNOWPLAYINGMETADATAKEY__ELAPSED FLOAT, '
          'Z_DKPERIODMETADATAKEY__PERIODEND TIMESTAMP, '
          'Z_DKPERIODMETADATAKEY__PERIODSTART TIMESTAMP, '
          'Z_QPMETRICSMETADATAKEY__TIMESTAMP FLOAT, '
          'Z_CDENTITYMETADATAKEY__BESTLANGUAGE VARCHAR, '
          'Z_CDENTITYMETADATAKEY__NAME VARCHAR, '
          'Z_CDPORTRAITMETADATAKEY__OSBUILD VARCHAR, '
          'Z_DKAPPINSTALLMETADATAKEY__PRIMARYCATEGORY VARCHAR, '
          'Z_DKAPPINSTALLMETADATAKEY__TITLE VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ACTIVITYTYPE VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__CONTENTDESCRIPTION VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ITEMIDENTIFIER VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ITEMRELATEDUNIQUEIDENTIFIER '
          'VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__SUGGESTEDINVOCATIONPHRASE '
          'VARCHAR, Z_DKAPPLICATIONACTIVITYMETADATAKEY__TITLE VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__USERACTIVITYREQUIREDSTRING '
          'VARCHAR, Z_DKAPPLICATIONACTIVITYMETADATAKEY__USERACTIVITYUUID '
          'VARCHAR, Z_DKAPPLICATIONMETADATAKEY__BACKBOARDSTATE VARCHAR, '
          'Z_DKAPPLICATIONMETADATAKEY__EXTENSIONCONTAININGBUNDLEIDENTIFIER '
          'VARCHAR, Z_DKAPPLICATIONMETADATAKEY__EXTENSIONHOSTIDENTIFIER '
          'VARCHAR, Z_DKAPPLICATIONMETADATAKEY__LAUNCHREASON VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__CHANNELS VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__DATASOURCES VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__IDENTIFIER VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__PORTNAME VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__PORTTYPE VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__PREFERREDDATASOURCE VARCHAR, '
          'Z_DKAUDIOMETADATAKEY__SELECTEDDATASOURCE VARCHAR, '
          'Z_DKBATTERYSAVERMETADATAKEY__SOURCE VARCHAR, '
          'Z_DKBLUETOOTHMETADATAKEY__ADDRESS VARCHAR, '
          'Z_DKBLUETOOTHMETADATAKEY__NAME VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__FEED VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__MESSAGE VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__SUBTITLE VARCHAR, '
          'Z_DKBULLETINBOARDMETADATAKEY__TITLE VARCHAR, '
          'Z_DKCALENDARMETADATAKEY__INTERACTION VARCHAR, '
          'Z_DKCALLMETADATAKEY__INTERACTION VARCHAR, '
          'Z_DKDEVICEIDMETADATAKEY__DEVICEIDENTIFIER VARCHAR, '
          'Z_DKDIGITALHEALTHMETADATAKEY__WEBDOMAIN VARCHAR, '
          'Z_DKINTENTMETADATAKEY__INTENTCLASS VARCHAR, '
          'Z_DKINTENTMETADATAKEY__INTENTVERB VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__URL VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__CITY VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__COUNTRY VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__DISPLAYNAME VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__FULLYFORMATTEDADDRESS '
          'VARCHAR, Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__LOCATIONNAME '
          'VARCHAR, Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__POSTALCODE_V2 '
          'VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__STATEORPROVINCE '
          'VARCHAR, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__SUBTHOROUGHFARE '
          'VARCHAR, Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__THOROUGHFARE '
          'VARCHAR, Z_DKLOCATIONMETADATAKEY__IDENTIFIER VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__HOMEUUID VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__VIEWINFORMATION VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__VIEWNAME VARCHAR, '
          'Z_DKMETADATAHOMEAPPVIEW__VIEWUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__ACCESSORYNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__ACCESSORYUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__CHARACTERISTICTYPE VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__CLIENTNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__HOMEUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__SERVICENAME VARCHAR, '
          'Z_DKMETADATAHOMEKITACCESSORYCONTROL__SERVICETYPE VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__ACTIONSETNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__ACTIONSETTYPE VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__ACTIONSETUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__CLIENTNAME VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__HOMEUUID VARCHAR, '
          'Z_DKMETADATAHOMEKITSCENE__SCENENAME VARCHAR, '
          'Z_DKMICROLOCATIONMETADATAKEY__LOCATIONDISTRIBUTION VARCHAR, '
          'Z_DKMICROLOCATIONMETADATAKEY__MICROLOCATIONDISTRIBUTION VARCHAR, '
          'Z_DKNOTIFICATIONUSAGEMETADATAKEY__BUNDLEID VARCHAR, '
          'Z_DKNOTIFICATIONUSAGEMETADATAKEY__IDENTIFIER VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__ALBUM VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__ARTIST VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__GENRE VARCHAR, '
          'Z_DKNOWPLAYINGMETADATAKEY__TITLE VARCHAR, '
          'Z_DKRELEVANTSHORTCUTMETADATAKEY__KEYIMAGEPROXYIDENTIFIER VARCHAR, '
          'Z_DKSAFARIHISTORYMETADATAKEY__TITLE VARCHAR, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__CLIENT VARCHAR, '
          'Z_DKSEARCHFEEDBACKMETADATAKEY__CONTACTID VARCHAR, '
          'Z_DKTOMBSTONEMETADATAKEY__EVENTSOURCEDEVICEID VARCHAR, '
          'Z_DKTOMBSTONEMETADATAKEY__EVENTSTREAMNAME VARCHAR, '
          'Z_QPMETRICSMETADATAKEY__QUERY VARCHAR, '
          'ZCOM_APPLE_CALENDARUIKIT_USERACTIVITY_EXTERNALID VARCHAR, '
          'ZKCDCSNOTIFICATIONOPTIONCLIENTIDENTIFIERKEY VARCHAR, '
          'ZKCDCSNOTIFICATIONOPTIONCLIENTLAUNCHKEY VARCHAR, '
          'ZKCDCSNOTIFICATIONOPTIONPERSISTENTPREDICATESTRINGKEY VARCHAR, '
          'ZMETADATAHASH VARCHAR, '
          'Z_DKAPPLICATIONACTIVITYMETADATAKEY__ITEMRELATEDCONTENTURL VARCHAR, '
          'Z_DKDIGITALHEALTHMETADATAKEY__WEBPAGEURL VARCHAR, '
          'Z_DKAPPINSTALLMETADATAKEY__SUBCATEGORIES BLOB, '
          'Z_DKINTENTMETADATAKEY__SERIALIZEDINTERACTION BLOB, '
          'Z_DKLOCATIONAPPLICATIONACTIVITYMETADATAKEY__PHONENUMBERS BLOB, '
          'Z_DKRELEVANTSHORTCUTMETADATAKEY__SERIALIZEDRELEVANTSHORTCUT BLOB, '
          'Z_QPMETRICSMETADATAKEY__CANDIDATELIST BLOB, '
          'Z_QPMETRICSMETADATAKEY__QUERYLIST BLOB )'),
      'ZSYNCPEER': (
          'CREATE TABLE ZSYNCPEER ( Z_PK INTEGER PRIMARY KEY, Z_ENT INTEGER, '
          'Z_OPT INTEGER, ZCLOUDID VARCHAR, ZDEVICEID VARCHAR, ZRAPPORTID '
          'VARCHAR, ZUUID BLOB )'),
      'Z_4EVENT': (
          'CREATE TABLE Z_4EVENT ( Z_4CUSTOMMETADATA INTEGER, Z_11EVENT '
          'INTEGER, PRIMARY KEY (Z_4CUSTOMMETADATA, Z_11EVENT) )'),
      'Z_METADATA': (
          'CREATE TABLE Z_METADATA (Z_VERSION INTEGER PRIMARY KEY, Z_UUID '
          'VARCHAR(255), Z_PLIST BLOB)'),
      'Z_MODELCACHE': (
          'CREATE TABLE Z_MODELCACHE (Z_CONTENT BLOB)'),
      'Z_PRIMARYKEY': (
          'CREATE TABLE Z_PRIMARYKEY (Z_ENT INTEGER PRIMARY KEY, Z_NAME '
          'VARCHAR, Z_SUPER INTEGER, Z_MAX INTEGER)')}

  SCHEMAS = [_SCHEMA_10_13, _SCHEMA_10_14]

  def KnowledgeCRow(
      self, parser_mediator, query, row, **unused_kwargs):
    """Parses KnowledgeC application activity

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    action = self._GetRowValue(query_hash, row, 'action')

    if action.startswith('/safari/'):
      event_data = MacKnowledgeCSafariEventData()
      event_data.url = self._GetRowValue(query_hash, row, 'zvaluestring')
      event_data.title = self._GetRowValue(query_hash, row, 'title')
    elif action.startswith('/app/'):
      event_data = MacKnowledgeCApplicationEventData()
      event_data.bundle_identifier = self._GetRowValue(
          query_hash, row, 'zvaluestring')
    else:
      # TODO: Add support for additional action types.
      return

    entry_creation = self._GetRowValue(query_hash, row, 'entry_creation')
    activity_starts = self._GetRowValue(query_hash, row, 'start')
    activity_ends = self._GetRowValue(query_hash, row, 'end')

    if activity_starts and activity_ends:
      event_data.duration = activity_ends - activity_starts

    entry_creation_time = dfdatetime_cocoa_time.CocoaTime(
        timestamp=entry_creation)
    entry_creation_event = time_events.DateTimeValuesEvent(
        entry_creation_time, definitions.TIME_DESCRIPTION_CREATION)
    activity_starts_time = dfdatetime_cocoa_time.CocoaTime(
        timestamp=activity_starts)
    parser_mediator.ProduceEventWithEventData(
        entry_creation_event, event_data)

    if activity_starts:
      activity_starts_event = time_events.DateTimeValuesEvent(
          activity_starts_time, definitions.TIME_DESCRIPTION_START)
      parser_mediator.ProduceEventWithEventData(
          activity_starts_event, event_data)

    if activity_ends:
      activity_ends_time = dfdatetime_cocoa_time.CocoaTime(
          timestamp=activity_ends)
      activity_ends_event = time_events.DateTimeValuesEvent(
          activity_ends_time, definitions.TIME_DESCRIPTION_END)
      parser_mediator.ProduceEventWithEventData(
          activity_ends_event, event_data)


sqlite.SQLiteParser.RegisterPlugin(MacKnowledgeCPlugin)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the syslog file event formatter."""

from __future__ import unicode_literals

import unittest

from plaso.formatters import syslog

from tests.formatters import test_lib


class SyslogLineFormatterTest(test_lib.EventFormatterTestCase):
  """Tests for the syslog line event formatter."""

  def testInitialization(self):
    """Tests the initialization."""
    event_formatter = syslog.SyslogLineFormatter()
    self.assertIsNotNone(event_formatter)

  def testGetFormatStringAttributeNames(self):
    """Tests the GetFormatStringAttributeNames function."""
    event_formatter = syslog.SyslogLineFormatter()

    expected_attribute_names = [
        'reporter',
        'pid',
        'body',
        'severity']

    self._TestGetFormatStringAttributeNames(
        event_formatter, expected_attribute_names)

    # TODO: add test for GetMessages.


class SyslogCommentFormatterTest(test_lib.EventFormatterTestCase):
  """Tests for the syslog comment formatter."""

  def testInitialization(self):
    """Tests the initialization."""
    event_formatter = syslog.SyslogCommentFormatter()
    self.assertIsNotNone(event_formatter)

  def testGetFormatStringAttributeNames(self):
    """Tests the GetFormatStringAttributeNames function."""
    event_formatter = syslog.SyslogCommentFormatter()

    expected_attribute_names = ['body']

    self._TestGetFormatStringAttributeNames(
        event_formatter, expected_attribute_names)

    # TODO: add test for GetMessages.


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Google Drive snapshots event formatter."""

from __future__ import unicode_literals

import unittest

from plaso.formatters import gdrive

from tests.formatters import test_lib


class GDriveCloudEntryFormatterTest(test_lib.EventFormatterTestCase):
  """Tests for the Google Drive snapshot cloud event formatter."""

  def testInitialization(self):
    """Tests the initialization."""
    event_formatter = gdrive.GDriveCloudEntryFormatter()
    self.assertIsNotNone(event_formatter)

  def testGetFormatStringAttributeNames(self):
    """Tests the GetFormatStringAttributeNames function."""
    event_formatter = gdrive.GDriveCloudEntryFormatter()

    expected_attribute_names = [
        'path', 'shared', 'size', 'url', 'document_type']

    self._TestGetFormatStringAttributeNames(
        event_formatter, expected_attribute_names)

  # TODO: add test for GetMessages.


class GDriveLocalEntryFormatterTest(test_lib.EventFormatterTestCase):
  """Tests for the Google Drive snapshot local event formatter."""

  def testInitialization(self):
    """Tests the initialization."""
    event_formatter = gdrive.GDriveLocalEntryFormatter()
    self.assertIsNotNone(event_formatter)

  def testGetFormatStringAttributeNames(self):
    """Tests the GetFormatStringAttributeNames function."""
    event_formatter = gdrive.GDriveLocalEntryFormatter()

    expected_attribute_names = ['path', 'size']

    self._TestGetFormatStringAttributeNames(
        event_formatter, expected_attribute_names)

  # TODO: add test for GetMessages.


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the file history ESE database event formatter."""

from __future__ import unicode_literals

import unittest

from plaso.formatters import file_history

from tests.formatters import test_lib


class FileHistoryNamespaceEventFormatterTest(
    test_lib.EventFormatterTestCase):
  """Tests for the file history ESE database event formatter."""

  def testInitialization(self):
    """Tests the initialization."""
    event_formatter = (
        file_history.FileHistoryNamespaceEventFormatter())
    self.assertIsNotNone(event_formatter)

  def testGetFormatStringAttributeNames(self):
    """Tests the GetFormatStringAttributeNames function."""
    event_formatter = (
        file_history.FileHistoryNamespaceEventFormatter())

    expected_attribute_names = [
        'original_filename', 'identifier', 'parent_identifier',
        'file_attribute', 'usn_number']

    self._TestGetFormatStringAttributeNames(
        event_formatter, expected_attribute_names)

  # TODO: add test for GetMessages.


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the MacOS Knowledge C db."""

from __future__ import unicode_literals

import unittest

from plaso.lib import definitions
from plaso.parsers.sqlite_plugins import mac_knowledgec

from tests.parsers.sqlite_plugins import test_lib


class MacKnowledgecTest(test_lib.SQLitePluginTestCase):
  """Tests for the MacOS KnowledgeC database."""

  def testProcessHighSierra(self):
    """Tests the Process function on a MacOS High Sierra database."""
    plugin = mac_knowledgec.MacKnowledgeCPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['mac_knowledgec-10.13.db'], plugin)

    self.assertEqual(0, storage_writer.number_of_warnings)
    self.assertEqual(51, storage_writer.number_of_events)
    events = list(storage_writer.GetEvents())
    event = events[0]
    self.CheckTimestamp(event.timestamp, '2019-02-10 16:59:58.860665')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_CREATION)

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(
        event_data.bundle_identifier, 'com.apple.Installer-Progress')

    expected_message = (
        'Application com.apple.Installer-Progress executed for 1 seconds')
    expected_short_message = 'Application com.apple.Installer-Progress'
    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

  def testProcessMojave(self):
    """Tests the Process function on a MacOS High Sierra database."""
    plugin = mac_knowledgec.MacKnowledgeCPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['mac_knowledgec-10.14.db'], plugin)

    self.assertEqual(0, storage_writer.number_of_warnings)
    self.assertEqual(231, storage_writer.number_of_events)
    events = list(storage_writer.GetEvents())

    event = events[225]
    self.CheckTimestamp(event.timestamp, '2019-05-08 13:57:30.668998')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_CREATION)

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.bundle_identifier, 'com.apple.Terminal')

    expected_message = (
        'Application com.apple.Terminal executed for 1041 seconds')
    expected_short_message = 'Application com.apple.Terminal'
    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

    event = events[212]
    self.CheckTimestamp(event.timestamp, '2019-05-08 13:57:20.000000')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_END)

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.url, 'https://www.instagram.com/')
    self.assertEqual(event_data.title, 'Instagram')

    expected_message = (
        'Visited: https://www.instagram.com/ (Instagram) Duration: 0')
    expected_short_message = 'Safari: https://www.instagram.com/'
    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the logging related classes and functions."""

from __future__ import unicode_literals

import logging
import os
import unittest

from plaso.lib import loggers

from tests import test_lib as shared_test_lib


class CompressedFileHandlerTests(shared_test_lib.BaseTestCase):
  """Tests for the compressed file handler for logging."""

  def testOpenAndEmit(self):
    """Tests the open and emit functions."""
    with shared_test_lib.TempDirectory() as temp_directory:
      filename = os.path.join(temp_directory, 'test.log.gz')
      handler = loggers.CompressedFileHandler(filename)

      record = logging.LogRecord(None, None, "", 0, "", (), None, None)
      handler.emit(record)


class LoggersTests(shared_test_lib.BaseTestCase):
  """Tests for the loggers module."""

  def testConfigureLogging(self):
    """Tests the ConfigureLogging function."""
    with shared_test_lib.TempDirectory() as temp_directory:
      filename = os.path.join(temp_directory, 'test.log.gz')
      loggers.ConfigureLogging(debug_output=True, filename=filename)

      filename = os.path.join(temp_directory, 'test.log')
      loggers.ConfigureLogging(filename=filename, quiet_mode=True)

    loggers.ConfigureLogging()


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the TLN output class."""

from __future__ import unicode_literals

import os
import unittest

from dfvfs.lib import definitions as dfvfs_definitions
from dfvfs.path import factory as path_spec_factory

from plaso.formatters import manager as formatters_manager
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.output import tln

from tests.cli import test_lib as cli_test_lib
from tests.containers import test_lib as containers_test_lib
from tests.output import test_lib


class TLNOutputModuleTest(test_lib.OutputModuleTestCase):
  """Tests for the TLN output module."""

  _OS_PATH_SPEC = path_spec_factory.Factory.NewPathSpec(
      dfvfs_definitions.TYPE_INDICATOR_OS, location='{0:s}{1:s}'.format(
          os.path.sep, os.path.join('cases', 'image.dd')))

  _TEST_EVENTS = [
      {'data_type': 'test:output',
       'display_name': 'OS: /var/log/syslog.1',
       'hostname': 'ubuntu',
       'inode': 12345678,
       'pathspec': path_spec_factory.Factory.NewPathSpec(
           dfvfs_definitions.TYPE_INDICATOR_TSK, inode=15,
           location='/var/log/syslog.1', parent=_OS_PATH_SPEC),
       'text': (
           'Reporter <CRON> PID: |8442| (pam_unix(cron:session): session\n '
           'closed for user root)'),
       'timestamp': timelib.Timestamp.CopyFromString('2012-06-27 18:17:01'),
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN,
       'username': 'root'}]

  def setUp(self):
    """Makes preparations before running an individual test."""
    self._output_writer = cli_test_lib.TestOutputWriter()
    output_mediator = self._CreateOutputMediator()
    self._output_module = tln.TLNOutputModule(output_mediator)
    self._output_module.SetOutputWriter(self._output_writer)

  def testWriteHeader(self):
    """Tests the WriteHeader function."""
    expected_header = 'Time|Source|Host|User|Description\n'

    self._output_module.WriteHeader()

    header = self._output_writer.ReadOutput()
    self.assertEqual(header, expected_header)

  def testWriteEventBody(self):
    """Tests the WriteEventBody function."""
    formatters_manager.FormattersManager.RegisterFormatter(
        test_lib.TestEventFormatter)

    event, event_data = containers_test_lib.CreateEventFromValues(
        self._TEST_EVENTS[0])
    self._output_module.WriteEventBody(event, event_data, None)

    expected_event_body = (
        '1340821021|LOG|ubuntu|root|2012-06-27T18:17:01+00:00; Unknown Time; '
        'Reporter <CRON> PID:  8442  (pam_unix(cron:session): '
        'session closed for user root)\n')

    event_body = self._output_writer.ReadOutput()
    self.assertEqual(event_body, expected_event_body)

    self.assertEqual(event_body.count('|'), 4)

    formatters_manager.FormattersManager.DeregisterFormatter(
        test_lib.TestEventFormatter)


class L2TTLNOutputModuleTest(test_lib.OutputModuleTestCase):
  """Tests for the log2timeline TLN output module."""

  _OS_PATH_SPEC = path_spec_factory.Factory.NewPathSpec(
      dfvfs_definitions.TYPE_INDICATOR_OS, location='{0:s}{1:s}'.format(
          os.path.sep, os.path.join('cases', 'image.dd')))

  _TEST_EVENTS = [
      {'data_type': 'test:output',
       'display_name': 'OS: /var/log/syslog.1',
       'hostname': 'ubuntu',
       'inode': 12345678,
       'pathspec': path_spec_factory.Factory.NewPathSpec(
           dfvfs_definitions.TYPE_INDICATOR_TSK, inode=15,
           location='/var/log/syslog.1', parent=_OS_PATH_SPEC),
       'text': (
           'Reporter <CRON> PID: |8442| (pam_unix(cron:session): session\n '
           'closed for user root)'),
       'timestamp': timelib.Timestamp.CopyFromString('2012-06-27 18:17:01'),
       'timestamp_desc': definitions.TIME_DESCRIPTION_UNKNOWN,
       'username': 'root'}]

  def setUp(self):
    """Makes preparations before running an individual test."""
    output_mediator = self._CreateOutputMediator()
    self._output_writer = cli_test_lib.TestOutputWriter()
    self._output_module = tln.L2TTLNOutputModule(output_mediator)
    self._output_module.SetOutputWriter(self._output_writer)

  def testWriteHeader(self):
    """Tests the WriteHeader function."""
    expected_header = 'Time|Source|Host|User|Description|TZ|Notes\n'

    self._output_module.WriteHeader()

    header = self._output_writer.ReadOutput()
    self.assertEqual(header, expected_header)

  def testWriteEventBody(self):
    """Tests the WriteEventBody function."""
    formatters_manager.FormattersManager.RegisterFormatter(
        test_lib.TestEventFormatter)

    event, event_data = containers_test_lib.CreateEventFromValues(
        self._TEST_EVENTS[0])
    self._output_module.WriteEventBody(event, event_data, None)

    expected_event_body = (
        '1340821021|LOG|ubuntu|root|2012-06-27T18:17:01+00:00; Unknown Time; '
        'Reporter <CRON> PID:  8442  (pam_unix(cron:session): '
        'session closed for user root)'
        '|UTC|File: OS: /var/log/syslog.1 inode: 12345678\n')

    event_body = self._output_writer.ReadOutput()
    self.assertEqual(event_body, expected_event_body)

    self.assertEqual(event_body.count('|'), 6)

    formatters_manager.FormattersManager.DeregisterFormatter(
        test_lib.TestEventFormatter)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the profiling CLI arguments helper."""

from __future__ import unicode_literals

import argparse
import unittest

from plaso.cli import tools
from plaso.cli.helpers import profiling
from plaso.lib import errors

from tests import test_lib as shared_test_lib
from tests.cli import test_lib as cli_test_lib


class ProfilingArgumentsHelperTest(cli_test_lib.CLIToolTestCase):
  """Tests for the profiling CLI arguments helper."""

  # pylint: disable=protected-access

  _EXPECTED_OUTPUT = """\
usage: cli_helper.py [--profilers PROFILERS_LIST]
                     [--profiling_directory DIRECTORY]
                     [--profiling_sample_rate SAMPLE_RATE]

Test argument parser.

optional arguments:
  --profilers PROFILERS_LIST
                        List of profilers to use by the tool. This is a comma
                        separated list where each entry is the name of a
                        profiler. Use "--profilers list" to list the available
                        profilers.
  --profiling_directory DIRECTORY, --profiling-directory DIRECTORY
                        Path to the directory that should be used to store the
                        profiling sample files. By default the sample files
                        are stored in the current working directory.
  --profiling_sample_rate SAMPLE_RATE, --profiling-sample-rate SAMPLE_RATE
                        Profiling sample rate (defaults to a sample every 1000
                        files).
"""

  def testAddArguments(self):
    """Tests the AddArguments function."""
    argument_parser = argparse.ArgumentParser(
        prog='cli_helper.py', description='Test argument parser.',
        add_help=False,
        formatter_class=cli_test_lib.SortedArgumentsHelpFormatter)

    profiling.ProfilingArgumentsHelper.AddArguments(argument_parser)

    output = self._RunArgparseFormatHelp(argument_parser)
    self.assertEqual(output, self._EXPECTED_OUTPUT)

  def testParseOptions(self):
    """Tests the ParseOptions function."""
    # pylint: disable=no-member

    test_tool = tools.CLITool()

    options = cli_test_lib.TestOptions()
    options.profiling_sample_rate = '100'

    profiling.ProfilingArgumentsHelper.ParseOptions(options, test_tool)
    self.assertEqual(test_tool._profiling_sample_rate, 100)

    with shared_test_lib.TempDirectory() as temp_directory:
      options = cli_test_lib.TestOptions()
      options.profilers = 'processing'
      options.profiling_directory = temp_directory

      profiling.ProfilingArgumentsHelper.ParseOptions(options, test_tool)
      self.assertEqual(test_tool._profilers, set(['processing']))
      self.assertEqual(test_tool._profiling_directory, temp_directory)
      self.assertEqual(test_tool._profiling_sample_rate, 1000)

    with self.assertRaises(errors.BadConfigObject):
      options = cli_test_lib.TestOptions()

      profiling.ProfilingArgumentsHelper.ParseOptions(options, None)

    with self.assertRaises(errors.BadConfigOption):
      options = cli_test_lib.TestOptions()
      options.profilers = 'bogus'

      profiling.ProfilingArgumentsHelper.ParseOptions(options, test_tool)

    with self.assertRaises(errors.BadConfigOption):
      options = cli_test_lib.TestOptions()
      options.profiling_directory = '/bogus'

      profiling.ProfilingArgumentsHelper.ParseOptions(options, test_tool)

    with self.assertRaises(errors.BadConfigOption):
      options = cli_test_lib.TestOptions()
      options.profiling_sample_rate = 'a'

      profiling.ProfilingArgumentsHelper.ParseOptions(options, test_tool)

    with self.assertRaises(errors.BadConfigOption):
      options = cli_test_lib.TestOptions()
      options.profiling_sample_rate = 100

      profiling.ProfilingArgumentsHelper.ParseOptions(options, test_tool)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The syslog cron formatters."""

from __future__ import unicode_literals

from plaso.formatters import interface
from plaso.formatters import manager


class CronTaskRunEventFormatter(interface.ConditionalEventFormatter):
  """Formatter for a syslog cron task run event."""

  DATA_TYPE = 'syslog:cron:task_run'

  FORMAT_STRING_SEPARATOR = ' '

  FORMAT_STRING_PIECES = [
      'Cron ran: {command}',
      'for user: {username}',
      'pid: {pid}']

  FORMAT_STRING_SHORT = '{body}'

  SOURCE_LONG = 'Cron log'
  SOURCE_SHORT = 'LOG'


manager.FormattersManager.RegisterFormatter(CronTaskRunEventFormatter)



# -*- coding: utf-8 -*-
"""A plugin that extracts browser history from events."""

from __future__ import unicode_literals

import collections
import re
import sys

if sys.version_info[0] < 3:
  import urllib as urlparse
else:
  from urllib import parse as urlparse # pylint: disable=no-name-in-module

# pylint: disable=wrong-import-position
from plaso.analysis import interface
from plaso.analysis import logger
from plaso.analysis import manager
from plaso.containers import reports
from plaso.lib import py2to3


# Create a lightweight object that is used to store timeline based information
# about each search term.
# pylint: disable=invalid-name
SEARCH_OBJECT = collections.namedtuple(
    'SEARCH_OBJECT', 'time source engine search_term')


class BrowserSearchPlugin(interface.AnalysisPlugin):
  """Analyze browser search entries from events."""

  NAME = 'browser_search'

  # Indicate that we do not want to run this plugin during regular extraction.
  ENABLE_IN_EXTRACTION = False

  _EVENT_TAG_COMMENT = 'Browser Search'
  _EVENT_TAG_LABELS = ['browser_search']

  _SUPPORTED_EVENT_DATA_TYPES = frozenset([
      'chrome:autofill:entry',
      'chrome:cache:entry',
      'chrome:cookie:entry',
      'chrome:extension_activity:activity_log',
      'chrome:history:file_downloaded',
      'chrome:history:page_visited',
      'firefox:cache:record',
      'firefox:cookie:entry',
      'firefox:places:bookmark_annotation',
      'firefox:places:bookmark_folder',
      'firefox:places:bookmark',
      'firefox:places:page_visited',
      'firefox:downloads:download',
      'cookie:google:analytics:utma',
      'cookie:google:analytics:utmb',
      'cookie:google:analytics:utmt',
      'cookie:google:analytics:utmz',
      'msiecf:leak',
      'msiecf:redirected',
      'msiecf:url',
      'msie:webcache:container',
      'msie:webcache:containers',
      'msie:webcache:leak_file',
      'msie:webcache:partitions',
      'opera:history:entry',
      'opera:history:typed_entry',
      'safari:cookie:entry',
      'safari:history:visit',
      'safari:history:visit_sqlite'])

  # TODO: use groups to build a single RE.

  # Here we define filters and callback methods for all hits on each filter.
  _URL_FILTERS = frozenset([
      ('Bing', re.compile(r'bing\.com/search'), '_ExtractSearchQueryFromURL'),
      ('DuckDuckGo', re.compile(r'duckduckgo\.com'),
       '_ExtractDuckDuckGoSearchQuery'),
      ('GMail', re.compile(r'mail\.google\.com'),
       '_ExtractGMailSearchQuery'),
      ('Google Docs', re.compile(r'docs\.google\.com'),
       '_ExtractGoogleDocsSearchQuery'),
      ('Google Drive', re.compile(r'drive\.google\.com/drive/search'),
       '_ExtractGoogleSearchQuery'),
      ('Google Search',
       re.compile(r'(www\.|encrypted\.|/)google\.[^/]*/search'),
       '_ExtractGoogleSearchQuery'),
      ('Google Sites', re.compile(r'sites\.google\.com/site'),
       '_ExtractGoogleSearchQuery'),
      ('Yahoo', re.compile(r'yahoo\.com/search'),
       '_ExtractYahooSearchQuery'),
      ('Yandex', re.compile(r'yandex\.com/search'),
       '_ExtractYandexSearchQuery'),
      ('Youtube', re.compile(r'youtube\.com'),
       '_ExtractYouTubeSearchQuery'),
  ])

  def __init__(self):
    """Initializes an analysis plugin."""
    super(BrowserSearchPlugin, self).__init__()
    self._counter = collections.Counter()

    # Store a list of search terms in a timeline format.
    # The format is key = timestamp, value = (source, engine, search term).
    self._search_term_timeline = []

  def _DecodeURL(self, url):
    """Decodes the URL, replaces %XX to their corresponding characters.

    Args:
      url (str): encoded URL.

    Returns:
      str: decoded URL.
    """
    if not url:
      return ''

    decoded_url = urlparse.unquote(url)
    if isinstance(decoded_url, py2to3.BYTES_TYPE):
      try:
        decoded_url = decoded_url.decode('utf-8')
      except UnicodeDecodeError as exception:
        decoded_url = decoded_url.decode('utf-8', errors='replace')
        logger.warning(
            'Unable to decode URL: {0:s} with error: {1!s}'.format(
                url, exception))

    return decoded_url

  def _ExtractDuckDuckGoSearchQuery(self, url):
    """Extracts a search query from a DuckDuckGo search URL.

    DuckDuckGo: https://duckduckgo.com/?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'q=' not in url:
      return None

    return self._GetBetweenQEqualsAndAmpersand(url).replace('+', ' ')

  def _ExtractGMailSearchQuery(self, url):
    """Extracts a search query from a GMail search URL.

    GMail: https://mail.google.com/mail/u/0/#search/query[/?]

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'search/' not in url:
      return None

    _, _, line = url.partition('search/')
    line, _, _ = line.partition('/')
    line, _, _ = line.partition('?')

    return line.replace('+', ' ')

  def _ExtractGoogleDocsSearchQuery(self, url):
    """Extracts a search query from a Google docs URL.

    Google Docs: https://docs.google.com/.*/u/0/?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query  or None if no query was found.
    """
    if 'q=' not in url:
      return None

    line = self._GetBetweenQEqualsAndAmpersand(url)
    if not line:
      return None

    return line.replace('+', ' ')

  def _ExtractGoogleSearchQuery(self, url):
    """Extracts a search query from a Google URL.

    Google Drive: https://drive.google.com/drive/search?q=query
    Google Search: https://www.google.com/search?q=query
    Google Sites: https://sites.google.com/site/.*/system/app/pages/
                  search?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'search' not in url or 'q=' not in url:
      return None

    line = self._GetBetweenQEqualsAndAmpersand(url)
    if not line:
      return None

    return line.replace('+', ' ')

  def _ExtractYahooSearchQuery(self, url):
    """Extracts a search query from a Yahoo search URL.

    Examples:
      https://search.yahoo.com/search?p=query
      https://search.yahoo.com/search;?p=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'p=' not in url:
      return None
    _, _, line = url.partition('p=')
    before_and, _, _ = line.partition('&')
    if not before_and:
      return None
    yahoo_search_url = before_and.split()[0]

    return yahoo_search_url.replace('+', ' ')

  def _ExtractYandexSearchQuery(self, url):
    """Extracts a search query from a Yandex search URL.

    Yandex: https://www.yandex.com/search/?text=query

    Args:
      url (str): URL.

    Returns:
      str: search query or None if no query was found.
    """
    if 'text=' not in url:
      return None
    _, _, line = url.partition('text=')
    before_and, _, _ = line.partition('&')
    if not before_and:
      return None
    yandex_search_url = before_and.split()[0]

    return yandex_search_url.replace('+', ' ')

  def _ExtractYouTubeSearchQuery(self, url):
    """Extracts a search query from a YouTube search URL.

    YouTube: https://www.youtube.com/results?search_query=query

    Args:
      url (str): URL.

    Returns:
      str: search query.
    """
    return self._ExtractSearchQueryFromURL(url)

  def _ExtractSearchQueryFromURL(self, url):
    """Extracts a search query from the URL.

    Bing: https://www.bing.com/search?q=query
    GitHub: https://github.com/search?q=query

    Args:
      url (str): URL.

    Returns:
      str: search query, the value between 'q=' and '&' or None if no
          query was found.
    """
    if 'search' not in url or 'q=' not in url:
      return None

    return self._GetBetweenQEqualsAndAmpersand(url).replace('+', ' ')

  def _GetBetweenQEqualsAndAmpersand(self, url):
    """Retrieves the substring between the substrings 'q=' and '&'.

    Args:
      url (str): URL.

    Returns:
      str: search query, the value between 'q=' and '&'  or None if no query
      was found.
    """
    # Make sure we're analyzing the query part of the URL.
    _, _, url = url.partition('?')
    # Look for a key value pair named 'q'.
    _, _, url = url.partition('q=')
    if not url:
      return ''

    # Strip additional key value pairs.
    url, _, _ = url.partition('&')
    return url

  def CompileReport(self, mediator):
    """Compiles an analysis report.

    Args:
      mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfvfs.

    Returns:
      AnalysisReport: analysis report.
    """
    results = {}
    for key, count in iter(self._counter.items()):
      search_engine, _, search_term = key.partition(':')
      results.setdefault(search_engine, {})
      results[search_engine][search_term] = count

    lines_of_text = []
    for search_engine, terms in sorted(results.items()):
      lines_of_text.append(' == ENGINE: {0:s} =='.format(search_engine))

      for search_term, count in sorted(
          terms.items(), key=lambda x: (x[1], x[0]), reverse=True):
        lines_of_text.append('{0:d} {1:s}'.format(count, search_term))

      # An empty string is added to have SetText create an empty line.
      lines_of_text.append('')

    lines_of_text.append('')
    report_text = '\n'.join(lines_of_text)
    analysis_report = reports.AnalysisReport(
        plugin_name=self.NAME, text=report_text)
    analysis_report.report_array = self._search_term_timeline
    analysis_report.report_dict = results
    return analysis_report

  def ExamineEvent(self, mediator, event, event_data):
    """Analyzes an event.

    Args:
      mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfvfs.
      event (EventObject): event.
      event_data (EventData): event data.
    """
    if event_data.data_type not in self._SUPPORTED_EVENT_DATA_TYPES:
      return

    url = getattr(event_data, 'url', None)
    if not url:
      return

    parser_or_plugin_name = getattr(event_data, 'parser', 'N/A')

    for engine, url_expression, method_name in self._URL_FILTERS:
      callback_method = getattr(self, method_name, None)
      if not callback_method:
        logger.warning('Missing method: {0:s}'.format(callback_method))
        continue

      match = url_expression.search(url)
      if not match:
        continue

      search_query = callback_method(url)
      if not search_query:
        logger.warning('Missing search query for URL: {0:s}'.format(url))
        continue

      search_query = self._DecodeURL(search_query)
      if not search_query:
        continue

      event_tag = self._CreateEventTag(
          event, self._EVENT_TAG_COMMENT, self._EVENT_TAG_LABELS)
      mediator.ProduceEventTag(event_tag)

      self._counter['{0:s}:{1:s}'.format(engine, search_query)] += 1

      # Add the timeline format for each search term.
      search_object = SEARCH_OBJECT(
          event.timestamp, parser_or_plugin_name, engine, search_query)
      self._search_term_timeline.append(search_object)


manager.AnalysisPluginManager.RegisterPlugin(BrowserSearchPlugin)



# -*- coding: utf-8 -*-
"""Apache access log (access.log) parser.

Parser based on the two default apache formats, common and combined log format
defined in https://httpd.apache.org/docs/2.4/logs.html
"""

from __future__ import unicode_literals

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements
from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import errors
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.parsers import manager
from plaso.parsers import text_parser


class ApacheAccessEventData(events.EventData):
  """Apache access event data.

  Attributes:
    http_request_referer (str): http request referer header information.
    http_request (str): first line of http request.
    http_request_user_agent (str): http request user agent header information.
    http_response_bytes (int): http response bytes size without headers.
    http_response_code (int): http response code from server.
    ip_address (str): IPv4 or IPv6 addresses.
    port_number (int): canonical port of the server serving the request.
    remote_name (str): remote logname (from identd, if supplied).
    server_name (str): canonical hostname of the server serving the request.
    user_name (str): logged user name.
  """

  DATA_TYPE = 'apache:access'

  def __init__(self):
    """Initializes event data."""
    super(ApacheAccessEventData, self).__init__(data_type=self.DATA_TYPE)
    self.http_request = None
    self.http_request_referer = None
    self.http_request_user_agent = None
    self.http_response_bytes = None
    self.http_response_code = None
    self.ip_address = None
    self.port_number = None
    self.remote_name = None
    self.server_name = None
    self.user_name = None


class ApacheAccessParser(text_parser.PyparsingSingleLineTextParser):
  """Apache access log file parser"""

  NAME = 'apache_access'
  DESCRIPTION = 'Apache access Parser'

  MAX_LINE_LENGTH = 2048

  # Date format [18/Sep/2011:19:18:28 -0400]
  _DATE_TIME = pyparsing.Group(
      pyparsing.Suppress('[') +
      text_parser.PyparsingConstants.TWO_DIGITS.setResultsName('day') +
      pyparsing.Suppress('/') +
      text_parser.PyparsingConstants.THREE_LETTERS.setResultsName('month') +
      pyparsing.Suppress('/') +
      text_parser.PyparsingConstants.FOUR_DIGITS.setResultsName('year') +
      pyparsing.Suppress(':') +
      text_parser.PyparsingConstants.TWO_DIGITS.setResultsName('hours') +
      pyparsing.Suppress(':') +
      text_parser.PyparsingConstants.TWO_DIGITS.setResultsName('minutes') +
      pyparsing.Suppress(':') +
      text_parser.PyparsingConstants.TWO_DIGITS.setResultsName('seconds') +
      pyparsing.Combine(
          pyparsing.oneOf(['-', '+']) +
          pyparsing.Word(
              pyparsing.nums, exact=4)).setResultsName('time_offset') +
      pyparsing.Suppress(']')).setResultsName('date_time')

  _HTTP_REQUEST = (
      pyparsing.Suppress('"') +
      pyparsing.SkipTo('"').setResultsName('http_request') +
      pyparsing.Suppress('"'))

  _PORT_NUMBER = text_parser.PyparsingConstants.INTEGER.setResultsName(
      'port_number')

  _REMOTE_NAME = (
      pyparsing.Word(pyparsing.alphanums) |
      pyparsing.Literal('-')).setResultsName('remote_name')

  _RESPONSE_BYTES = (
      pyparsing.Literal('-') |
      text_parser.PyparsingConstants.INTEGER).setResultsName('response_bytes')

  _REFERER = (
      pyparsing.Suppress('"') +
      pyparsing.SkipTo('"').setResultsName('referer') +
      pyparsing.Suppress('"'))

  _SERVER_NAME = (
      pyparsing.Word(pyparsing.alphanums + '-' + '.').setResultsName(
          'server_name'))

  _USER_AGENT = (
      pyparsing.Suppress('"') +
      pyparsing.SkipTo('"').setResultsName('user_agent') +
      pyparsing.Suppress('"'))

  _USER_NAME = (
      pyparsing.Word(pyparsing.alphanums) |
      pyparsing.Literal('-')).setResultsName('user_name')

  # Defined in https://httpd.apache.org/docs/2.4/logs.html
  # format: "%h %l %u %t \"%r\" %>s %b"
  _COMMON_LOG_FORMAT_LINE = (
      text_parser.PyparsingConstants.IP_ADDRESS.setResultsName('ip_address') +
      _REMOTE_NAME +
      _USER_NAME +
      _DATE_TIME +
      _HTTP_REQUEST +
      text_parser.PyparsingConstants.INTEGER.setResultsName('response_code') +
      _RESPONSE_BYTES +
      pyparsing.lineEnd())

  # Defined in https://httpd.apache.org/docs/2.4/logs.html
  # format: "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\""
  _COMBINED_LOG_FORMAT_LINE = (
      text_parser.PyparsingConstants.IP_ADDRESS.setResultsName('ip_address') +
      _REMOTE_NAME +
      _USER_NAME +
      _DATE_TIME +
      _HTTP_REQUEST +
      text_parser.PyparsingConstants.INTEGER.setResultsName('response_code') +
      _RESPONSE_BYTES +
      _REFERER +
      _USER_AGENT +
      pyparsing.lineEnd())

  # "vhost_combined" format as used by Debian and related distributions.
  # "%v:%p %h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\""
  _VHOST_COMBINED_LOG_FORMAT = (
      _SERVER_NAME +
      pyparsing.Suppress(':') +
      _PORT_NUMBER +
      text_parser.PyparsingConstants.IP_ADDRESS.setResultsName('ip_address') +
      _REMOTE_NAME +
      _USER_NAME +
      _DATE_TIME +
      _HTTP_REQUEST +
      text_parser.PyparsingConstants.INTEGER.setResultsName('response_code') +
      _RESPONSE_BYTES +
      _REFERER +
      _USER_AGENT +
      pyparsing.lineEnd()
  )

  LINE_STRUCTURES = [
      ('combined_log_format', _COMBINED_LOG_FORMAT_LINE),
      ('common_log_format', _COMMON_LOG_FORMAT_LINE),
      ('vhost_combined_log_format', _VHOST_COMBINED_LOG_FORMAT)]

  _SUPPORTED_KEYS = frozenset([key for key, _ in LINE_STRUCTURES])

  # TODO: migrate function after dfdatetime issue #47 is fixed.
  def _GetISO8601String(self, structure):
    """Normalize date time parsed format to an ISO 8601 date time string.
    The date and time values in Apache access log files are formatted as:
    "[18/Sep/2011:19:18:28 -0400]".

    Args:
      structure (pyparsing.ParseResults): structure of tokens derived from a
          line of a text file.

    Returns:
      str: ISO 8601 date time string.

    Raises:
      ValueError: if the structure cannot be converted into a date time string.
    """
    month = self._GetValueFromStructure(structure, 'month')

    try:
      month = timelib.MONTH_DICT.get(month.lower(), 0)
    except AttributeError as exception:
      raise ValueError('unable to parse month with error: {0!s}.'.format(
          exception))

    time_offset = self._GetValueFromStructure(structure, 'time_offset')

    try:
      time_offset_hours = int(time_offset[1:3], 10)
      time_offset_minutes = int(time_offset[3:5], 10)
    except (IndexError, TypeError, ValueError) as exception:
      raise ValueError(
          'unable to parse time zone offset with error: {0!s}.'.format(
              exception))

    year = self._GetValueFromStructure(structure, 'year')
    day_of_month = self._GetValueFromStructure(structure, 'day')
    hours = self._GetValueFromStructure(structure, 'hours')
    minutes = self._GetValueFromStructure(structure, 'minutes')
    seconds = self._GetValueFromStructure(structure, 'seconds')

    try:
      date_time_string = (
          '{0:04d}-{1:02d}-{2:02d}T{3:02d}:{4:02d}:{5:02d}.000000'
          '{6:s}{7:02d}:{8:02d}').format(
              year, month, day_of_month, hours, minutes, seconds,
              time_offset[0], time_offset_hours, time_offset_minutes)
    except ValueError as exception:
      raise ValueError(
          'unable to format date time string with error: {0!s}.'.format(
              exception))

    return date_time_string

  def ParseRecord(self, parser_mediator, key, structure):
    """Parses a matching entry.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
        and other components, such as storage and dfvfs.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): elements parsed from the file.

    Raises:
      ParseError: when the structure type is unknown.
    """
    if key not in self._SUPPORTED_KEYS:
      raise errors.ParseError(
          'Unable to parse record, unknown structure: {0:s}'.format(key))

    date_time = dfdatetime_time_elements.TimeElements()

    date_time_string = self._GetValueFromStructure(structure, 'date_time')

    try:
      iso_date_time = self._GetISO8601String(date_time_string)
      date_time.CopyFromStringISO8601(iso_date_time)
    except ValueError:
      parser_mediator.ProduceExtractionWarning(
          'invalid date time value: {0!s}'.format(date_time_string))
      return

    event = time_events.DateTimeValuesEvent(
        date_time, definitions.TIME_DESCRIPTION_RECORDED)

    event_data = ApacheAccessEventData()
    event_data.ip_address = self._GetValueFromStructure(structure, 'ip_address')
    event_data.remote_name = self._GetValueFromStructure(
        structure, 'remote_name')
    event_data.user_name = self._GetValueFromStructure(structure, 'user_name')
    event_data.http_request = self._GetValueFromStructure(
        structure, 'http_request')
    event_data.http_response_code = self._GetValueFromStructure(
        structure, 'response_code')
    event_data.http_response_bytes = self._GetValueFromStructure(
        structure, 'response_bytes')

    if key in ('combined_log_format', 'vhost_combined_log_format'):
      event_data.http_request_referer = self._GetValueFromStructure(
          structure, 'referer')
      event_data.http_request_user_agent = self._GetValueFromStructure(
          structure, 'user_agent')

    if key == 'vhost_combined_log_format':
      event_data.server_name = self._GetValueFromStructure(
          structure, 'server_name')
      event_data.port_number = self._GetValueFromStructure(
          structure, 'port_number')

    parser_mediator.ProduceEventWithEventData(event, event_data)

  # pylint: disable=unused-argument
  def VerifyStructure(self, parser_mediator, line):
    """Verifies that this is an apache access log file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
        and other components, such as storage and dfvfs.
      line (str): line from the text file.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    return max([parser.matches(line) for _, parser in self.LINE_STRUCTURES])


manager.ParsersManager.RegisterParser(ApacheAccessParser)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Keychain password database parser."""

from __future__ import unicode_literals

import unittest

from plaso.formatters import mac_keychain as _  # pylint: disable=unused-import
from plaso.lib import definitions
from plaso.parsers import mac_keychain

from tests.parsers import test_lib


class MacKeychainParserTest(test_lib.ParserTestCase):
  """Tests for keychain file parser."""

  def testParse(self):
    """Tests the Parse function."""
    parser = mac_keychain.KeychainParser()
    storage_writer = self._ParseFile(['login.keychain'], parser)

    self.assertEqual(storage_writer.number_of_warnings, 0)
    self.assertEqual(storage_writer.number_of_events, 8)

    events = list(storage_writer.GetEvents())

    event = events[0]

    self.CheckTimestamp(event.timestamp, '2014-01-26 14:51:48.000000')
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_CREATION)

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.entry_name, 'Secret Application')
    self.assertEqual(event_data.account_name, 'moxilo')
    expected_ssgp = (
        'b8e44863af1cb0785b89681d22e2721997ccfb8adb8853e726aff94c8830b05a')
    self.assertEqual(event_data.ssgp_hash, expected_ssgp)
    expected_message = 'Name: Secret Application Account: moxilo'
    expected_short_message = 'Secret Application'
    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

    event = events[1]

    self.assertEqual(
        event.timestamp_desc,
        definitions.TIME_DESCRIPTION_MODIFICATION)

    self.CheckTimestamp(event.timestamp, '2014-01-26 14:52:29.000000')

    event = events[2]

    self.CheckTimestamp(event.timestamp, '2014-01-26 14:53:29.000000')

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.entry_name, 'Secret Note')
    self.assertEqual(event_data.text_description, 'secure note')
    self.assertEqual(len(event_data.ssgp_hash), 1696)
    expected_message = 'Name: Secret Note'
    expected_short_message = 'Secret Note'
    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)

    event = events[4]

    self.CheckTimestamp(event.timestamp, '2014-01-26 14:54:33.000000')

    event_data = self._GetEventDataOfEvent(storage_writer, event)
    self.assertEqual(event_data.entry_name, 'plaso.kiddaland.net')
    self.assertEqual(event_data.account_name, 'MrMoreno')
    expected_ssgp = (
        '83ccacf55a8cb656d340ec405e9d8b308fac54bb79c5c9b0219bd0d700c3c521')
    self.assertEqual(event_data.ssgp_hash, expected_ssgp)
    self.assertEqual(event_data.where, 'plaso.kiddaland.net')
    self.assertEqual(event_data.protocol, 'http')
    self.assertEqual(event_data.type_protocol, 'dflt')

    expected_message = (
        'Name: plaso.kiddaland.net '
        'Account: MrMoreno '
        'Where: plaso.kiddaland.net '
        'Protocol: http (dflt)')
    expected_short_message = 'plaso.kiddaland.net'
    self._TestGetMessageStrings(
        event_data, expected_message, expected_short_message)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Compound ZIP plugin related functions and classes for testing."""

from __future__ import unicode_literals

from tests.parsers import test_lib


class CompoundZIPPluginTestCase(test_lib.ParserTestCase):
  """The unit test case for a compound ZIP plugin."""



# -*- coding: utf-8 -*-
"""The Windows Recycler/Recycle Bin formatter."""

from __future__ import unicode_literals

from plaso.formatters import interface
from plaso.formatters import manager
from plaso.lib import errors


class WinRecyclerFormatter(interface.ConditionalEventFormatter):
  """Formatter for a Windows Recycler/Recycle Bin file event."""

  DATA_TYPE = 'windows:metadata:deleted_item'

  _DRIVE_LETTER = {
      0x00: 'A',
      0x01: 'B',
      0x02: 'C',
      0x03: 'D',
      0x04: 'E',
      0x05: 'F',
      0x06: 'G',
      0x07: 'H',
      0x08: 'I',
      0x09: 'J',
      0x0A: 'K',
      0x0B: 'L',
      0x0C: 'M',
      0x0D: 'N',
      0x0E: 'O',
      0x0F: 'P',
      0x10: 'Q',
      0x11: 'R',
      0x12: 'S',
      0x13: 'T',
      0x14: 'U',
      0x15: 'V',
      0x16: 'W',
      0x17: 'X',
      0x18: 'Y',
      0x19: 'Z',
  }

  # The format string.
  FORMAT_STRING_PIECES = [
      'DC{record_index} ->',
      '{original_filename}',
      '[{short_filename}]',
      '(from drive: {drive_letter})']

  FORMAT_STRING_SHORT_PIECES = [
      'Deleted file: {original_filename}']

  SOURCE_LONG = 'Recycle Bin'
  SOURCE_SHORT = 'RECBIN'

  # pylint: disable=unused-argument
  def GetMessages(self, formatter_mediator, event_data):
    """Determines the formatted message strings for the event data.

    Args:
      formatter_mediator (FormatterMediator): mediates the interactions
          between formatters and other components, such as storage and Windows
          EventLog resources.
      event_data (EventData): event data.

    Returns:
      tuple(str, str): formatted message string and short message string.

    Raises:
      WrongFormatter: if the event data cannot be formatted by the formatter.
    """
    if self.DATA_TYPE != event_data.data_type:
      raise errors.WrongFormatter('Unsupported data type: {0:s}.'.format(
          event_data.data_type))

    event_values = event_data.CopyToDict()

    drive_number = event_values.get('drive_number', None)
    event_values['drive_letter'] = self._DRIVE_LETTER.get(
        drive_number, 'UNKNOWN')

    return self._ConditionalFormatMessages(event_values)


manager.FormattersManager.RegisterFormatter(WinRecyclerFormatter)
